from ..transformer import BPtTransformerMV
from ..input import MVTransformer
from ....dataset.Dataset import Dataset
from ....main.input import Pipeline, Scaler, Model, ParamSearch
from ....main.funcs import get_estimator
import tempfile
import shutil
import os
import numpy as np
from mvlearn.embed import CCA


def get_X():

    X = np.array([[0., 0., 1., 0.1, -0.2],
                  [1., 0., 0., 0.9, 1.1],
                  [2., 2., 2., 6.2, 5.9],
                  [3., 5., 4., 11.9, 12.3]])

    return X


def basic_test():

    X = get_X()
    mv = BPtTransformerMV(estimator=CCA(multiview_output=False),
                          inds=[[0, 1, 2], [3, 4]])

    X_trans = mv.fit_transform(X)

    # Basic Checks
    assert X_trans.shape == (4, 1)
    assert mv.inds_ == [0, 1, 2, 3, 4]
    assert mv.view_inds_ == [[0, 1, 2], [3, 4]]
    assert mv.out_mapping_[0] == [0]
    assert mv.out_mapping_[4] == [0]
    assert len(mv.out_mapping_) == 5
    assert isinstance(mv.estimator_, CCA)
    assert mv.n_trans_feats_ == 1
    assert mv.n_features_in_ == 5

    assert len(mv.estimator_.means_) == 2
    assert len(mv.estimator_.means_[0] == 3)
    assert len(mv.estimator_.means_[1] == 2)

    assert len(mv.estimator_.loadings_) == 2
    assert len(mv.estimator_.loadings_[0]) == 3
    assert len(mv.estimator_.loadings_[1]) == 2


def test_with_cache():

    temp_dr = os.path.join(tempfile.gettempdir(), 'temp_dr')
    if os.path.exists(temp_dr):
        shutil.rmtree(temp_dr)

    X = get_X()
    mv = BPtTransformerMV(estimator=CCA(multiview_output=False),
                          inds=[[0, 1, 2], [3, 4]], cache_loc=temp_dr)

    # Fit once w/ caching
    mv.fit_transform(X)
    assert os.listdir(temp_dr) == ['joblib']

    # Fit again, should load from cached
    X_trans = mv.fit_transform(X)
    assert X_trans.shape == (4, 1)

    if os.path.exists(temp_dr):
        shutil.rmtree(temp_dr)

    assert not os.path.exists(temp_dr)


def test_with_mapping():

    X = get_X()
    mv = BPtTransformerMV(estimator=CCA(multiview_output=False),
                          inds=[[0, 1, 2], [3, 4]])

    # This mapping should ignore feat 1, adding it on after
    mapping = {0: 0, 1: 0, 2: 2, 3: 3, 4: 4}
    X_trans = mv.fit_transform(X, mapping=mapping)

    assert X_trans.shape == (4, 2)
    assert np.array_equal(X_trans[:, 1], X[:, 1])

    assert len(mv.estimator_.loadings_) == 2
    assert len(mv.estimator_.loadings_[0]) == 2
    assert len(mv.estimator_.loadings_[1]) == 2
    assert len(mv.estimator_.means_) == 2
    assert len(mv.estimator_.means_[0] == 2)
    assert len(mv.estimator_.means_[1] == 2)


def get_dataset():

    data = Dataset()
    data['0'] = [1, 0, 1, .3, .5, 7]
    data['1'] = [0, 0, 1, .1, -1, 1]
    data['2'] = [0, 1, 0, .2, -2, 2.2]
    data['3'] = [2, 2, 2, 0, 2, 0]
    data['4'] = [0, 1, 0, 4.3, 4.3, 10]
    data['5'] = [0, 0, 1, 0, 1, 0]

    data = data.add_scope(['0', '1', '2'], 'c1')
    data = data.add_scope(['3', '4'], 'c2')
    data = data.set_target('5')
    return data


def test_build():

    data = get_dataset()

    piece = MVTransformer(obj='cca', scopes=[['1', '2'], ['3', '4']])
    obj, params = piece.build(data)
    trans = obj[1]
    assert isinstance(trans, BPtTransformerMV)
    assert len(trans.inds) == 2
    assert trans.inds[0] == [1, 2]
    assert trans.inds[1] == [3, 4]

    cca = trans.estimator
    assert isinstance(cca, CCA)

    # Should be empty
    assert len(params) == 0

    # Try fit
    x_trans = trans.fit_transform(np.array(data[['0', '1', '2', '3', '4']]))
    assert x_trans.shape == (6, 2)
    assert np.array_equal(x_trans[:, 1], np.array(data['0']))


def test_basic_pipeline_integration():

    data = get_dataset()
    mv_trans = MVTransformer(obj='cca',
                             scopes=[['1', '2'], ['3', '4']],
                             n_components=2)

    pipe = Pipeline([Scaler('standard'), mv_trans, Model('linear')])
    est = get_estimator(pipe, data)

    assert len(est.steps) == 3
    assert isinstance(est.steps[1][1], BPtTransformerMV)
    assert isinstance(est.steps[1][1].estimator, CCA)

    # Get data
    X, y = data.get_Xy()

    # Fit
    est.fit(X, y)

    pred_y = est.predict(X)
    assert len(pred_y) == 6

    # Just make sure no errors scoring
    est.score(X, y)

    # Transform shape should be 2 comp + feat 0
    X_trans = est.transform(X)
    assert X_trans.shape == (6, 3)


def test_pipeline_with_search_integration():
    '''Want to make sure doesn't break in a multi-proc context.'''

    data = get_dataset()
    mv_trans = MVTransformer(obj='cca',
                             scopes=[['1', '2'], ['3', '4']],
                             n_components=2)

    param_search = ParamSearch(n_iter=2)
    pipe = Pipeline([Scaler('standard'), mv_trans, Model('dt', params=1)],
                    param_search=param_search)
    search_est = get_estimator(pipe, data, problem_type='regression', n_jobs=2)

    # Some assertions about estimator
    est = search_est.estimator
    assert len(est.steps) == 3
    assert isinstance(est.steps[1][1], BPtTransformerMV)
    assert isinstance(est.steps[1][1].estimator, CCA)

    # Get data
    X, y = data.get_Xy()

    # Fit
    search_est.fit(X, y)

    pred_y = search_est.predict(X)
    assert len(pred_y) == 6

    # Just make sure no errors scoring
    search_est.score(X, y)

    # Transform shape should be 2 comp + feat 0
    X_trans = search_est.transform(X)
    assert X_trans.shape == (6, 3)



from ..helpers import (to_memmap, from_memmap, get_grid_params,
                       is_ng, update_mapping)
import numpy as np
import os
from ...default.params.Params import Choice, TransitionChoice, Scalar
import pytest
import nevergrad as ng


def test_to_memmap():

    X = np.ones((10, 10), dtype='float32')
    f, dtype, shape = to_memmap(X)

    assert os.path.exists(f)
    assert dtype == 'float32'
    assert shape == X.shape

    os.unlink(f)
    assert not os.path.exists(f)


def test_from_memmap():

    X = np.ones((10, 10), dtype='float32')
    X_m = to_memmap(X)
    X_c = from_memmap(X_m)

    assert X_c.shape == X.shape
    assert X.dtype == X_c.dtype

    X_c[:] = 0
    X_c = from_memmap(X_m)
    assert np.all(X == 1)


def test_get_grid_params():

    params = {'1': Choice([1, 2, 3]),
              '2': TransitionChoice([1, 2]),
              '3': 3,
              '4': Scalar(lower=1, upper=2).set_integer_casting()}

    grid_params = get_grid_params(params)

    assert grid_params['1'] == [1, 2, 3]
    assert grid_params['2'] == [1, 2]
    assert grid_params['3'] == 3
    assert grid_params['4'] == [1, 2]


def test_get_grid_params_fail():

    with pytest.raises(RuntimeError):

        params = {'1': ng.p.Scalar(lower=1, upper=2)}
        get_grid_params(params)


def test_is_ng():

    assert not is_ng(6)
    assert not is_ng([1, 2, 3])
    assert not is_ng(set([5]))
    assert not is_ng(from_memmap)
    assert not is_ng(np.ones(10))
    assert is_ng(ng.p.Choice([1, 2]))
    assert is_ng(ng.p.TransitionChoice([1, 2]))
    assert is_ng(ng.p.Scalar())
    assert is_ng(ng.p.Log(init=1))


def test_update_mapping_base():

    mapping = {1: 1}
    new_mapping = {}

    update_mapping(mapping, new_mapping)
    assert mapping[1] == 1
    assert len(mapping) == 1


def test_update_mapping_simple():

    mapping = {0: 0, 1: 1}
    new_mapping = {0: 1, 1: 0}
    update_mapping(mapping, new_mapping)

    assert len(mapping) == 2
    assert mapping[0] == 1
    assert mapping[1] == 0


def test_update_mapping_lists():

    mapping = {0: [0, 1], 1: [2, 3]}
    new_mapping = {1: [11], 2: [22]}
    update_mapping(mapping, new_mapping)

    # Sorts
    assert mapping[0] == [0, 11]
    assert mapping[1] == [3, 22]
    assert len(mapping) == 2


def test_update_mapping_none():

    mapping = {0: None}
    new_mapping = {0: 1}
    update_mapping(mapping, new_mapping)

    assert mapping[0] is None
    assert len(mapping) == 1



from sklearn.base import BaseEstimator, TransformerMixin
import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import OneHotEncoder
import pandas as pd


class LinearResidualizer(BaseEstimator, TransformerMixin):
    '''This class is used for residualizing data with the LinearRegression
    model from scikit-learn according to a DataFrame with one or
    more columns.

    In the case of feature columns with any NaN's, the linear model
    will be fit just on all subject's with non NaN data, and NaN's
    propagated.

    Warning: you like should not use this transformer to
    transform categorical data.

    Parameters
    ------------
    to_resid_df : pandas DataFrame or BPt Dataset
        Should contain one or more columns as indexed
        by the same index as provided input features, this
        df will be used to residualize the input data.

        There may not be any NaN data here, and there must
        be data for each subject (i.e., any subject
        potentially present in X, must have a value here.)

    demean : bool, optional
        If True, then the variables passed in to_resid_df
        will be de-meaned according to the actual group
        of subjects passed during fit. For transform,
        values will be de-meaned according to the mean
        for each column as calculated during fit.

        If dummy_code is True, then dummy coded variables
        will be de-meaned after dummy coding.

        ::

            default = True

    dummy_code : bool, optional
        If True, then any categorical variables,
        as determined by checking to see if the underlying
        pandas dtype is 'category', will be dummy coded
        with scikit-learn's :class:`OneHotEncoder`.

        If no categorical variables, this option will be skipped.

        If demean is True, then categorical variables will
        first be dummy coded and then de-meaned.

        ::

            default = True

    fit_intercept : bool, optional
        If True, then the intercept of the
        linear model will be fitted. It will then
        be added onto the residualized data as
        intercept + difference in real - predicted

        ::

            default = False
    '''

    _needs_fit_index = True
    _needs_transform_index = True

    def __init__(self, to_resid_df, demean=True,
                 dummy_code=True, fit_intercept=True):

        self.to_resid_df = to_resid_df
        self.demean = demean
        self.dummy_code = dummy_code
        self.fit_intercept = fit_intercept

        self._check_args()

    def _transform_covars(self, X, index=None, fit=True):

        # Grab covars as copy of indexed df
        covars = self.to_resid_df.loc[index].copy()

        if self.dummy_code:

            # Find which cols are categorical
            cat_cols = [col for col in covars
                        if covars[col].dtype.name == 'category']

            # Must be more than 0 cat columns
            if len(cat_cols) > 0:

                # Cast to np
                cat_trans = np.array(covars[cat_cols])

                # Init and fit if fit
                if fit:
                    self.encoder_ = OneHotEncoder(drop='first', sparse=False)
                    self.encoder_.fit(cat_trans)

                # Transform
                cat_trans = self.encoder_.transform(cat_trans)

                # Stack with non-cat in place
                rest_cols = [col for col in covars if col not in cat_cols]
                covars = np.hstack([np.array(covars[rest_cols]), cat_trans])

        # At this stage make sure np array if not already
        covars = np.array(covars)

        if self.demean:

            # Need to calculate means if fit
            if fit:
                self.means_ = np.mean(covars, axis=0)

            covars = covars - self.means_

        return covars

    def fit(self, X, y=None, fit_index=None):

        if isinstance(X, pd.DataFrame):
            fit_index = X.index
            X = np.array(X)

        # Get covars as proc'ed np array, fitting demean or dummy code.
        covars = self._transform_covars(X, index=fit_index, fit=True)
        return self._fit(X, covars)

    def transform(self, X, transform_index=None):

        if isinstance(X, pd.DataFrame):
            transform_index = X.index
            X = np.array(X)

        # Get covars as proc'ed np array
        covars =\
            self._transform_covars(X, index=transform_index, fit=False)

        # Returned transformed copy
        return self._transform(X, covars)

    def _check_args(self):

        if not isinstance(self.to_resid_df, pd.DataFrame):
            raise RuntimeError('to_resid_df must be a DataFrame or Dataset.')

        if not isinstance(self.demean, bool):
            raise RuntimeError('demean must be True or False.')

        if not isinstance(self.dummy_code, bool):
            raise RuntimeError('dummy code must be True or False.')

        if not isinstance(self.fit_intercept, bool):
            raise RuntimeError('fit_intercept must be True or False.')

    def _fit(self, X, covars):

        self._check_args()

        # If NaN, use NaN fit
        if np.isnan(X).any():
            return self._nan_fit(X, covars)

        # Fit linear regression
        self.estimator_ = LinearRegression(fit_intercept=self.fit_intercept)
        self.estimator_.fit(covars, X)

        # Save score
        self.score_ = self.estimator_.score(covars, X)

        return self

    def _nan_fit(self, X, covars):

        # For each feature seperately
        self.estimators_ = []
        for i in range(X.shape[1]):

            # Operate on non-nan subjects for this feature
            mask = ~np.isnan(X[:, i])

            # If not at least 2 subjects valid,
            # skip and propegate NaN.
            if np.sum(mask) > 1:

                # Fit model
                model = LinearRegression(fit_intercept=self.fit_intercept)
                model.fit(covars[mask], X[mask, i])
                self.estimators_.append(model)

            # During transform, if skipped here, then will
            # transform to all NaNs.
            else:
                self.estimators_.append(None)

        return self

    def _transform(self, X, covars):

        # if fit with NaN's
        if hasattr(self, 'estimators_'):
            return self._nan_transform(X, covars)

        # Otherwise use estimator
        # The difference is the real value
        # minus the predicted value
        dif = X - self.estimator_.predict(covars)

        # Set resid as either diff or diff + intercept
        resid = dif

        if self.fit_intercept:
            resid += self.estimator_.intercept_

        return resid

    def _nan_transform(self, X, covars):

        # Init empty resid array of NaN's
        resid = np.empty(shape=X.shape)
        resid[:] = np.nan

        # For each feature seperately
        for i in range(X.shape[1]):

            # Operate on non-nan subjects for this feature
            mask = ~np.isnan(X[:, i])

            # If not at least 2 subjects valid,
            # skip and propegate NaN.
            if len(mask) > 1:

                # Grab saved estimator
                model = self.estimators_[i]

                # Compute difference of real value - predicted
                dif_i = X[mask, i] - model.predict(covars[mask])

                # Set resid as either diff or diff + intercept
                resid_i = dif_i

                if self.fit_intercept:
                    resid_i += model.intercept_

                # Fill in NaN mask
                resid[mask, i] = resid_i

        return resid

    def fit_transform(self, X, y=None, fit_index=None):

        if isinstance(X, pd.DataFrame):
            fit_index = X.index
            X = np.array(X)

        covars = self._transform_covars(X, index=fit_index, fit=True)
        return self._fit(X, covars)._transform(X, covars)

    def __repr__(self):

        with pd.option_context('display.max_rows', 1,
                               'display.max_columns', 1,
                               'display.max_colwidth', 1):
            rep = super().__repr__()
        return rep



from ..main.CV import CVStrategy
from .helpers import save_subjects
import pandas as pd
from .Dataset import _shared_docs
from pandas.util._decorators import doc

_val_docs = {}

_val_docs['size'] = '''size : float, int or None, optional
        | This parameter represents the size of the train/test split
          to apply. If passed a floating type, this parameter
          should be between 0.0 and 1.0, and it will specify the
          proportion / percent of the data to bet set in the train/test split.
          If passed as an integer type, this value will be intrpetted as
          the absolute number of subjects to be set in the train/test split.

        | Note that either this parameter or subjects should be used, not both
          as they define different behaviors. Keep as default = None if using
          subjects.

        ::

            default = None
'''

_val_docs['subjects'] = '''subjects : :ref:`Subjects`, optional
        | This parameter can be optionally used instead of size in
          the case that a specific set of subjects should be used to define
          the split. This argument can accept any valid :ref:`Subjects` style
          input. Explicitly, either this parameter or size should be used,
          not both as they define different behaviors.

        | In the case that additional subjects are specified here, i.e., ones
          not loaded in the current Dataset, they will be simply be ignored,
          and the functional splits set to the overlap of passed subjects
          with loaded subjects.
          valid subjects.

        ::

            default = None
'''

_val_docs['cv_strategy'] = '''cv_strategy : None or :class:`CVStrategy`, optional
        This parameter is only relevant when size is used, i.e., a new split is
        defined (and subjects is not used). In this case,
        an instance of :class:`CVStrategy` defining any validation
        behavior the train/test split should be performed according
        to should be passed - or left as None (the default), which
        will use random splits. This parameter is typically
        used to define behavior like making sure the same distribution
        of target variable is present in both folds, or that
        members of the same family are preserved across folds.

        ::

            default = None
'''

_val_docs['random_state'] = '''random_state : int or None, optional
        This parameter is only relevant when size is used, i.e., a new split is
        defined (and subjects is not used). In this case, this
        parameter represents the random state in which the split should
        be performed according to. Random states allow for
        reproducing the same train/test splits across different
        runs if given the same input Dataset. If left as
        None, the train/test split will be performed with
        just a random random seed (that is to say a different random state
        each time the function is called.)

        ::

            default = None
'''


def _validate_cv_key(self, cv_key, name):
    '''Various input validation.'''

    if cv_key not in list(self):
        raise KeyError(f'Passed {cv_key} must be a specific loaded column')

    # Make sure categorical - scopes should already be checked
    if not self._is_category(cv_key, check_scopes=False):
        raise RuntimeError('Passed ' + name + ': ' + cv_key + ' must '
                           'be of type category! This can be set '
                           'in a number of ways, e.g., via '
                           'binarize, ordinalize, or directly via '
                           'add_scope.')

    # Check to see if any NaN's in this requested column
    nan_subjects = self._get_nan_subjects(col=cv_key)
    if len(nan_subjects) > 0:
        raise RuntimeError('Requested ' + name + ' col: ' + cv_key + ' '
                           'has ' + str(len(nan_subjects)) + ' NaN values.'
                           ' It cannot have any!')


def _proc_cv_strategy(self, cv_params):
    '''This function accepts cv_params and returns
    a CV object.'''

    # Check scopes and roles
    self._check_sr()

    # If None, return base
    if cv_params is None:
        return CVStrategy()

    # @TODO Support scikit-learn style CV directly.

    # If already a CV object, return as is
    if isinstance(cv_params, CVStrategy):
        return cv_params

    # Load train_only_subjects as sorted numpy array
    train_only_subjects = self.get_subjects(cv_params.train_only_subjects,
                                            return_as='flat index')

    # @TODO verbose statement saying X number of train_only_subjects loaded,
    # and also how many of those actually overlap with loaded subjects, and
    # a warning about how non-overlapping ones will be ignored.

    # Set to None if None
    if len(train_only_subjects) == 0:
        train_only_subjects = None

    # Unpack
    group_key = cv_params.groups
    strat_key = cv_params.stratify

    # If groups
    if group_key is not None:

        # Overlapping validations
        self._validate_cv_key(cv_key=group_key, name='groups')

        # Make sure it is non input
        if self.roles[group_key] != 'non input':
            raise RuntimeError('Passed groups: ' + group_key + ' must have '
                               'role = non input. This can be set with '
                               'set_role(...)')

        # Return CV
        return CVStrategy(groups=self[group_key],
                          train_only=train_only_subjects)

    # If stratify
    elif strat_key is not None:

        # Overlapping validations
        self._validate_cv_key(cv_key=strat_key, name='stratify')

        # Make sure it is non input or target
        if self.roles[strat_key] != 'non input' and \
           self.roles[strat_key] != 'target':

            raise RuntimeError('Passed stratify: ' + strat_key + ' must have '
                               'role = non input or target. This can be set '
                               'with set_role(...)')

        # Return CV
        return CVStrategy(stratify=self[strat_key],
                          train_only=train_only_subjects)

    # If None
    return CVStrategy(train_only=train_only_subjects)


def _validate_split(self, size, subjects):

    self._check_test_subjects()
    self._check_train_subjects()

    if self.train_subjects is not None:
        self._print('Overriding existing train/test split.', level=1)

    if size is None and subjects is None:
        raise TypeError('Either size or subjects must be non-null.')
    elif size is not None and subjects is not None:
        raise TypeError('Only size or subjects can be passed, not both.')


def _finish_split(self):

    # Save in class, pd.Index style
    self.train_subjects = pd.Index(self.train_subjects,
                                   name=self.index.name)
    self.test_subjects = pd.Index(self.test_subjects,
                                  name=self.index.name)

    self._print('Performed train/test split', level=1)
    self._print('Train size:', len(self.train_subjects), level=1)
    self._print('Test size: ', len(self.test_subjects), level=1)


@doc(**_val_docs, inplace=_shared_docs['inplace'])
def set_test_split(self, size=None, subjects=None,
                   cv_strategy=None, random_state=None, inplace=False):
    '''| Defines a set of subjects to be reserved as test subjects. This
         method includes utilities for either defining a new test split,
         or loading an existing one.

    | This method applies the passed parameters in order define a test set
      which is stored in the current Dataset.

    Parameters
    ----------
    {size}

    {subjects}

    {cv_strategy}

    {random_state}

    {inplace}

    See Also
    ----------
    set_train_split : Apply a train/test split but via specifying which
        subjects are training subjects.
    test_split : Apply a test split returning two separate Train and Test
        Datasets.
    save_test_split : Save the test subjects from a split to a text file.

    Examples
    ---------

    .. ipython:: python

        import BPt as bp

        # Load example data
        data = bp.read_pickle('data/example1.dataset')
        data

        data.set_test_split(size=.6, inplace=True)
        data.train_subjects
        data.test_subjects

    Note that the split is stored in the dataset itself.
    We can also pass specific subjects.

    .. ipython:: python

        data = data.set_test_split(subjects=[0, 1])
        data.train_subjects
        data.test_subjects

    '''

    if not inplace:
        return self._inplace('set_test_split', locals())

    # Input validation
    self._validate_split(size, subjects)

    # If size is passed
    if size is not None:

        # Check for if size 0
        if size == 0:
            self.test_subjects =\
                self.get_subjects(None, return_as='index')

            # If multi index, set as flat index
            if isinstance(self.index, pd.MultiIndex):
                self.train_subjects = self.index.to_flat_index()
            else:
                self.train_subjects = self.index

            self._print('Warning: Test size of 0 passed, all subjects will be',
                        'considered train subjects.', level=0)

        # Otherwise perform split according to size, cv and random_state
        else:

            # Process the passed cv params
            cv_obj = self._proc_cv_strategy(cv_strategy)

            # Get the overlap of train only subjects from loaded data
            _, valid_subjects, train_only_subjects =\
                cv_obj.get_train_only(self.index)

            # Print info on split
            self._print('Performing test split on:', len(valid_subjects),
                        'subjects', end='', level=1)

            if len(train_only_subjects) > 0:
                self._print(' with:', len(train_only_subjects), 'set as',
                            'train only', end='', level=1)
            self._print('.', level=1)

            self._print('random_state:', random_state, level=1)
            self._print('Test split size:', size, level=1)

            self.train_subjects, self.test_subjects = cv_obj.train_test_split(
                subjects=self.index, test_size=size,
                random_state=random_state, return_index=False)

        self._print(level=1)

    # If subjects is passed
    else:

        # Load all test subjects
        self.test_subjects = self.get_subjects(subjects, return_as='set')

        # Take only the overlap of the passed subjects with what is loaded
        self.test_subjects = [subject for subject in self.test_subjects
                              if subject in self.index]

        # Set remaining subjects to train subjects
        self.train_subjects = [subject for subject in self.index
                               if subject not in self.test_subjects]

    self._finish_split()


@doc(**_val_docs, inplace=_shared_docs['inplace'])
def set_train_split(self, size=None, subjects=None,
                    cv_strategy=None, random_state=None, inplace=False):
    '''| Defines a set of subjects to be reserved as train subjects. This
         method includes utilities for either defining a new train split,
         or loading an existing one.

    | This method applies the passed parameters in order define a train set
      which is stored in the current Dataset.

    Parameters
    ----------
    {size}

    {subjects}

    {cv_strategy}

    {random_state}

    {inplace}

    See Also
    ----------
    set_test_split : Apply a train/test split but via specifying which
        subjects are test subjects.
    train_split : Apply a train split returning two separate Train and Test
        Datasets.
    save_train_split : Save the train subjects from a split to a text file.

    Examples
    ---------

    .. ipython:: python

        import BPt as bp

        # Load example data
        data = bp.read_pickle('data/example1.dataset')
        data

        data.set_train_split(size=.6, inplace=True)
        data.train_subjects
        data.test_subjects

    Note that the split is stored in the dataset itself.
    We can also pass specific subjects.

    .. ipython:: python

        data = data.set_train_split(subjects=[0, 1])
        data.train_subjects
        data.test_subjects

    '''

    if not inplace:
        return self._inplace('set_train_split', locals())

    # Input validation
    self._validate_split(size, subjects)

    # If size is passed
    if size is not None:

        # Check for if size 0
        if size == 0:
            raise RuntimeError('Train size of 0 not allowed!')

        # Otherwise perform split according to size, cv and random_state
        else:

            # Process the passed cv params
            cv_obj = self._proc_cv_strategy(cv_strategy)

            # Get the overlap of train only subjects from loaded data
            _, valid_subjects, train_only_subjects =\
                cv_obj.get_train_only(self.index)

            # Print info on split
            self._print('Performing train split on:', len(valid_subjects),
                        'subjects', end='', level=1)

            if len(train_only_subjects) > 0:
                self._print(' with:', len(train_only_subjects), 'set as',
                            'train only', end='', level=1)
            self._print('.', level=1)

            self._print('random_state:', random_state, level=1)
            self._print('Train split size:', size, level=1)

            # Convert size to test size
            if isinstance(size, int):
                test_size = len(valid_subjects) - size
            else:
                test_size = 1 - size

            self.train_subjects, self.test_subjects = cv_obj.train_test_split(
                subjects=self.index, test_size=test_size,
                random_state=random_state, return_index=False)

        self._print(level=1)

    # If subjects is passed
    else:

        # Load all train subjects
        self.train_subjects = self.get_subjects(subjects, return_as='set')

        # Take only the overlap of the passed subjects with what is loaded
        self.train_subjects = [subject for subject in self.train_subjects
                               if subject in self.index]

        # Set remaining subjects to test subjects
        self.test_subjects = [subject for subject in self.index
                              if subject not in self.train_subjects]

    self._finish_split()


@doc(**_val_docs)
def test_split(self, size=None, subjects=None,
               cv_strategy=None, random_state=None):
    '''| This method defines and returns a Train and Test Dataset
         based on the passed parameters, allowing for defining
         the test set as a new split or from an existing set of
         subjects.

    | This method's parameters describe how the test set should be generated,
      where the training set is then defined by just which subjects are not in
      the test set.

    Parameters
    ----------
    {size}

    {subjects}

    {cv_strategy}

    {random_state}

    Returns
    -------
    train_data : :class:`Dataset`
        The current :class:`Dataset` as indexed (i.e., only the subjects from)
        the requested training set, as defined by the passed parameters.
        This :class:`Dataset` will have all the same metadata as the original
        :class:`Dataset`, though if changes are made to it, they will not
        influence the original :class:`Dataset`.

    test_data : :class:`Dataset`
        The current :class:`Dataset` as indexed (i.e., only the subjects from)
        the requested test set, as defined by the passed parameters.
        This :class:`Dataset` will have all the same metadata as the original
        :class:`Dataset`, though if changes are made to it, they will not
        influence the original :class:`Dataset`.

    See Also
    ----------
    train_split : Return a train/test split but via specifying which
        subjects are training subjects.
    set_test_split : Apply a test split, but storing the split information
        in the Dataset.
    save_test_split : Save the test subjects from a split to a text file.

    Examples
    ---------

    .. ipython:: python

        import BPt as bp

        # Load example data
        data = bp.read_pickle('data/example1.dataset')
        data

        tr_data, test_data = data.test_split(size=.2)
        tr_data
        test_data

    We can also define a split by passing specific subjects.

    .. ipython:: python

        tr_data, test_data = data.test_split(subjects=[3, 4])
        tr_data
        test_data

    We see that the parameters are used to define which subjects
    are set as test subjects.

    '''

    # Apply split on copy of data
    data = self.set_test_split(size=size, subjects=subjects,
                               cv_strategy=cv_strategy,
                               random_state=random_state,
                               inplace=False)

    # Properly extracts splits as separate Datasets
    return split(data)


@doc(**_val_docs)
def train_split(self, size=None, subjects=None,
                cv_strategy=None, random_state=None):
    '''| This method defines and returns a Train and Test Dataset
         based on the passed parameters, allowing for defining
         the training set as a new split or from an existing set of
         subjects.

    | This method's parameters describe how the training set
      should be generated, where the testing set is then defined by
      just which subjects are not in the training set.

    Parameters
    ----------
    {size}

    {subjects}

    {cv_strategy}

    {random_state}

    Returns
    -------
    train_data : :class:`Dataset`
        The current :class:`Dataset` as indexed (i.e., only the subjects from)
        the requested training set, as defined by the passed parameters.
        This :class:`Dataset` will have all the same metadata as the original
        :class:`Dataset`, though if changes are made to it, they will not
        influence the original :class:`Dataset`.

    test_data : :class:`Dataset`
        The current :class:`Dataset` as indexed (i.e., only the subjects from)
        the requested test set, as defined by the passed parameters.
        This :class:`Dataset` will have all the same metadata as the original
        :class:`Dataset`, though if changes are made to it, they will not
        influence the original :class:`Dataset`.

    See Also
    ----------
    test_split : Return a train/test split but via specifying which
        subjects are test subjects.
    set_train_split : Apply a train split, but storing the split information
        in the Dataset.
    save_train_split : Save the train subjects from a split to a text file.

    Examples
    ---------

    .. ipython:: python

        import BPt as bp

        # Load example data
        data = bp.read_pickle('data/example1.dataset')
        data

        tr_data, test_data = data.train_split(size=.6)
        tr_data
        test_data

        tr_data, test_data = data.train_split(subjects=[0, 1])
        tr_data
        test_data

    '''

    # Apply split on copy of data
    data = self.set_train_split(size=size, subjects=subjects,
                                cv_strategy=cv_strategy,
                                random_state=random_state,
                                inplace=False)

    # Properly extracts splits as separate Datasets
    return split(data)


def save_test_split(self, loc):
    '''Saves the currently defined test
    subjects in a text file with one subject / index
    per line.

    Parameters
    ----------
    loc : str or Path
        The location in which to save the test subjects
    '''

    self._check_test_subjects()

    if self.test_subjects is None:
        raise RuntimeError('No train test split defined')

    save_subjects(loc, self.test_subjects)


def save_train_split(self, loc):
    '''Saves the currently defined train
    subjects in a text file with one subject / index
    per line.

    Parameters
    ----------
    loc : str or Path
        The location in which to save the train subjects
    '''

    self._check_train_subjects()

    if self.train_subjects is None:
        raise RuntimeError('No train test split defined')

    save_subjects(loc, self.train_subjects)


def split(data):

    # Get the train and test subjects
    train_subjects = data.get_subjects('train')
    test_subjects = data.get_subjects('test')

    # Reset split to None
    data.train_subjects, data.test_subjects = None, None

    # Get as copies - since meta data should be separate
    tr_data = data.loc[train_subjects].copy(deep=False)
    test_data = data.loc[test_subjects].copy(deep=False)

    # Return Datasets as train then test
    return tr_data, test_data



from .BPtPipeline import BPtPipeline
from ..main.input import Custom
from ..util import is_array_like
from .BPtSearchCV import get_search_cv


def _check_for_user_passed(objs, cnt, up):

    if objs is not None:

        # If list / array like passed
        if is_array_like(objs):

            # Call recursively on each entry
            for o in range(len(objs)):
                objs[o], cnt = _check_for_user_passed(objs[o], cnt, up)

        # If a single obj
        else:

            if hasattr(objs, 'base_model'):
                if objs.base_model is not None:
                    objs.base_model, cnt =\
                        _check_for_user_passed(objs.base_model, cnt, up)
            if hasattr(objs, 'models'):
                if objs.models is not None:
                    objs.models, cnt =\
                        _check_for_user_passed(objs.models, cnt, up)
            if hasattr(objs, 'target_scaler'):
                if objs.target_scaler is not None:
                    objs.target_scaler, cnt =\
                        _check_for_user_passed(objs.target_scaler, cnt, up)

            # If a Param obj - call recursively to set the value of the
            # base obj
            if hasattr(objs, 'obj'):
                objs.obj, cnt = _check_for_user_passed(objs.obj, cnt, up)

            # Skip custom
            elif isinstance(objs, Custom):
                pass

            # Now, we assume any single obj that gets here, if not
            # a str is user passed custom obj
            elif not isinstance(objs, str):
                save_name = 'Custom ' + str(cnt)
                cnt += 1
                up[save_name] = objs
                objs = save_name

    return objs, cnt


def check_for_user_passed(objs):

    user_passed = {}
    objs, _ = _check_for_user_passed(objs, 0, up=user_passed)
    return objs, user_passed


class BPtPipelineConstructor():

    def __init__(self, pipeline_params, problem_spec, dataset):

        # Save param search here - should be None or dict
        self.param_search = pipeline_params.param_search

        # Extract params for BPtPipeline
        self.cache_loc = pipeline_params.cache_loc
        self.verbose = pipeline_params.verbose

        # Save some params to pass around when building the steps
        spec = problem_spec._get_spec()

        # Get params as ordered list of steps
        pipe_params = pipeline_params._get_steps()

        # Create the pipeline pieces
        self._create_pipeline_pieces(
            pipe_params=pipe_params,
            dataset=dataset, spec=spec)

    def _create_pipeline_pieces(self, pipe_params,
                                dataset, spec):

        # First check for user passed - update in place
        # Check for user passed out here to avoid possible name collisions.
        pipe_params, user_passed_objs = check_for_user_passed(pipe_params)

        # Want to process all pieces with the same constructor
        # at the same time, then put back in the right order

        # Create empty list of objects to fill
        self.objs = [None for _ in range(len(pipe_params))]

        # Params is dict, so unordered
        self.params = {}

        # Get unique constructors
        constructors = [param._constructor for param in pipe_params]
        unique = set(constructors)

        # For each unique piece_type
        for constr_type in unique:

            # Get a list of the pieces of this type
            # and their original index
            in_scope, inds = [], []
            for i in range(len(pipe_params)):
                if constructors[i] is constr_type:
                    in_scope.append(pipe_params[i])
                    inds.append(i)

            # Init the correct constructor
            constructor = constr_type(spec=spec, dataset=dataset,
                                      user_passed_objs=user_passed_objs)

            # Process the pipe params of this type
            objs, params = constructor.process(in_scope)

            # Put in correct spots
            for i, ind in enumerate(inds):
                self.objs[ind] = objs[i]

            # Update params
            self.params.update(params)

        # Now self.objs contains the initialized steps, and self.params
        # the corresponding param dists.

    def get_pipeline(self):
        '''Make the model pipeline object'''

        pipeline = BPtPipeline(self.objs, verbose=self.verbose,
                               cache_loc=self.cache_loc)

        return pipeline

    def is_search(self):

        if self.param_search is None:
            return False
        return True

    def get_search_wrapped_pipeline(self):

        # Grab the base pipeline
        base_pipeline = self.get_pipeline()

        # If no search, just return copy of pipeline
        # and the params seperately
        if not self.is_search():
            return base_pipeline, self.params

        # Create the search object
        search_model = get_search_cv(estimator=base_pipeline,
                                     param_search=self.param_search,
                                     param_distributions=self.params)

        # Return search model, and empty dict, since params are used
        return search_model, {}


def get_pipe(pipeline_params, problem_spec, dataset):

    # Init the Pipeline, which creates the pipeline pieces
    pipeline_constructor =\
        BPtPipelineConstructor(pipeline_params=pipeline_params,
                               problem_spec=problem_spec,
                               dataset=dataset)
    # Set the final model // search wrap
    Model, pipeline_params =\
        pipeline_constructor.get_search_wrapped_pipeline()

    return Model, pipeline_params



from ..loaders import (SurfMaps, SurfLabels, Identity,
                       load_surf, SingleConnectivityMeasure,
                       ThresholdNetworkMeasures)
import numpy as np
import tempfile
import os
import pytest
import warnings


def test_identity():

    i = Identity()
    x = np.ones((10, 10))
    assert i.fit_transform(x).shape == (100, )


def test_load_surf_from_np():

    tmp = os.path.join(tempfile.gettempdir(), 'fake.npy')
    np.save(tmp, np.ones(10))

    loaded = load_surf(tmp)
    assert loaded.shape == (10,)
    os.remove(tmp)


def test_load_surf_None():

    loaded = load_surf(None)
    assert loaded is None


def test_load_surf_bad():

    with pytest.raises(FileNotFoundError):
        load_surf('not_real.npy')


def test_load_nilearn():

    try:
        import nibabel as nib
    except ImportError:
        return

    temp_loc = os.path.join(tempfile.gettempdir(), 'temp.nii')
    obj = nib.Nifti1Image(np.ones((100, 1, 1, 1)), np.eye(4))
    nib.save(obj, temp_loc)

    warnings.simplefilter('ignore')
    loaded = load_surf(temp_loc)
    assert loaded.shape == (100,)


def test_load_surf_array():

    loaded = load_surf(np.ones(5))
    assert np.sum(loaded) == 5


def test_surf_labels():

    sl = SurfLabels(labels=np.array([1, 1, 2, 2]))
    data = np.array([5, 5, 10, 10])

    Xt = sl.fit_transform(data)
    assert np.array_equal(Xt, np.array([5, 10]))

    reverse_trans = sl.inverse_transform(Xt)
    assert np.array_equal(reverse_trans, np.array([5, 5, 10, 10]))


def test_surf_labels2():

    sl = SurfLabels(labels=np.array([0, 2, 2, 2, 3, 3]))
    data = np.array([1, 2, 3, 4, 5, 6])

    Xt = sl.fit_transform(data)
    assert np.array_equal(Xt, np.array([3, 5.5]))

    reverse_trans = sl.inverse_transform(Xt)
    assert np.array_equal(reverse_trans, np.array([0, 3, 3, 3, 5.5, 5.5]))


def test_surf_labels_mask():

    labels = np.array([1, 1, 2, 2])
    data = np.array([1, 2, 3, 4])

    mask = np.array([1, 0, 0, 0])
    Xt = SurfLabels(labels=labels, mask=mask,
                    strategy='mean').fit_transform(data)

    assert np.array_equal(Xt, np.array([2, 3.5]))


def test_bad_surf_labels_mask():

    labels = np.array([1, 1, 2, 2])
    data = np.array([1, 2, 3, 4])
    mask = np.array([1, 0, 0, 0, 0])

    with pytest.raises(RuntimeError):
        SurfLabels(labels=labels, mask=mask,
                   strategy='mean').fit_transform(data)


def test_surf_labels_dif_strats():

    labels = np.array([1, 1, 2, 2])
    data = np.array([1, 2, 3, 4])

    Xt = SurfLabels(labels=labels,
                    strategy='median').fit_transform(data)
    assert np.array_equal(Xt, np.array([1.5, 3.5]))

    Xt = SurfLabels(labels=labels,
                    strategy='sum').fit_transform(data)
    assert np.array_equal(Xt, np.array([3, 7]))

    Xt = SurfLabels(labels=labels,
                    strategy='min').fit_transform(data)
    assert np.array_equal(Xt, np.array([1, 3]))

    Xt = SurfLabels(labels=labels,
                    strategy='max').fit_transform(data)
    assert np.array_equal(Xt, np.array([2, 4]))

    Xt = SurfLabels(labels=labels,
                    strategy='std').fit_transform(data)
    assert np.array_equal(Xt, np.array([.5, .5]))

    Xt = SurfLabels(labels=labels,
                    strategy='var').fit_transform(data)
    assert np.array_equal(Xt, np.array([.25, .25]))


def test_surf_data_2d():

    labels = np.array([1, 2, 1, 2])
    data = np.array([[1, 2, 3], [2, 3, 4],
                     [1, 2, 3], [2, 3, 4]])

    sl = SurfLabels(labels=labels)
    Xt = sl.fit_transform(data)
    assert np.array_equal(Xt, np.array([1, 2, 3, 2, 3, 4]))

    reverse_trans = sl.inverse_transform(Xt)
    assert np.array_equal(reverse_trans, np.array([[1, 2, 3],
                                                   [2, 3, 4],
                                                   [1, 2, 3],
                                                   [2, 3, 4]]))


def test_surf_data_2d_no_vectorize():

    labels = np.array([1, 2, 1, 2])
    data = np.array([[1, 2, 3], [2, 3, 4],
                     [1, 2, 3], [2, 3, 4]])

    sl = SurfLabels(labels=labels, vectorize=False)
    Xt = sl.fit_transform(data)
    assert np.array_equal(Xt, np.array([[1, 2, 3], [2, 3, 4]]))

    reverse_trans = sl.inverse_transform(Xt)
    assert np.array_equal(reverse_trans, np.array([[1, 2, 3],
                                                   [2, 3, 4],
                                                   [1, 2, 3],
                                                   [2, 3, 4]]))


def test_surf_labels_2d_fail():

    labels = np.array([[1, 1, 1], [2, 2, 2], [1, 1, 1], [2, 2, 2]])
    data = np.array([[1, 2, 3], [2, 3, 4], [1, 2, 3], [2, 3, 4]])

    sl = SurfLabels(labels=labels, vectorize=False)

    with pytest.raises(RuntimeError):
        sl.fit_transform(data)


def test_surf_maps_ls():

    maps = np.array([[0, 1],
                     [0, 2],
                     [1, 0],
                     [1, 0]])

    data = np.array([1.0, 1.0, 5.0, 5.0])
    data_dtype = data.dtype.name

    sm = SurfMaps(maps=maps, strategy='ls')
    Xt = sm.fit_transform(data)

    assert data_dtype == Xt.dtype.name
    np.testing.assert_almost_equal(Xt, np.array([5, .6]), decimal=6)

    # Reverse transform
    r_trans = sm.inverse_transform(Xt)
    np.array_equal(r_trans, np.array([.6, 1.2, 5, 5]))


def test_surf_maps_average():

    maps = np.array([[1, 0],
                     [1, 0],
                     [0, 1],
                     [0, 2]])

    data = np.array([1.0, 3.0, 1.0, 2.0])
    data_dtype = data.dtype.name

    sm = SurfMaps(maps=maps, strategy='average')
    Xt = sm.fit_transform(data)

    assert data_dtype == Xt.dtype.name
    np.array_equal(Xt, np.array([2, 1.66666667]))


def test_surf_maps_average_fail():

    maps = np.array([[1, 0],
                     [1, 0],
                     [0, 1],
                     [0, 2]])

    data = np.array([1.0, 3.0, 1.0, 2.0])

    sm = SurfMaps(maps=maps, strategy='average')
    Xt = sm.fit_transform(data)

    with pytest.raises(RuntimeError):
        sm.inverse_transform(Xt)


def test_single_connectivity_measure():

    scm = SingleConnectivityMeasure()
    X_trans = scm.fit_transform(np.ones((5, 10)))
    assert X_trans.shape == (10, 10)

    scm = SingleConnectivityMeasure(vectorize=True, discard_diagonal=True)
    X_trans = scm.fit_transform(np.ones((5, 10)))
    assert X_trans.shape == (45, )

    scm = SingleConnectivityMeasure(vectorize=True)
    X_trans = scm.fit_transform(np.ones((5, 10)))
    assert X_trans.shape == (55, )


def test_threshold_network_measures_abs():

    nm =\
        ThresholdNetworkMeasures(threshold=0.2,
                                 threshold_type='abs',
                                 threshold_method='value',
                                 to_compute='avg_degree')

    X = np.array([[.1, .2, .3, -.5],
                  [.2, 0, -.5, 0],
                  [.3, -.5, 0, 0],
                  [-.5, 0, 0, 0]])

    # Fit
    nm.fit(X)
    assert nm._feat_names[0] == 'avg_degree'
    assert len(nm._feat_names) == 1

    # Threshold should stay fixed
    nm._threshold_check(X)
    assert nm.threshold == .2

    thresh_X = nm._apply_threshold(X)
    assert len(np.unique(thresh_X)) == 2

    # Should just drop .1 edge
    assert np.sum(thresh_X) == 8

    import networkx as nx
    G = nx.from_numpy_matrix(thresh_X)

    degrees = [n[1] for n in G.degree()]
    X_trans = nm.transform(X)
    assert len(X_trans) == 1
    assert np.mean(degrees) == X_trans[0]


def test_threshold_network_measures_neg():

    nm =\
        ThresholdNetworkMeasures(threshold=-.2,
                                 threshold_type='neg',
                                 threshold_method='value')

    X = np.array([[.1, .2, .3, -.5],
                  [.2, 0, -.5, 0],
                  [.3, -.5, 0, 0],
                  [-.5, 0, 0, 0]])

    thresh_X = nm._apply_threshold(X)
    assert np.sum(thresh_X) == 4
    assert thresh_X[0][-1] == 1
    assert thresh_X[0][1] == 0


def test_threshold_network_measures_pos():

    nm =\
        ThresholdNetworkMeasures(threshold=.2,
                                 threshold_type='pos',
                                 threshold_method='value')

    X = np.array([[.1, .2, .3, -.5],
                  [.2, 0, -.5, 0],
                  [.3, -.5, 0, 0],
                  [-.5, 0, 0, 0]])

    thresh_X = nm._apply_threshold(X)
    assert np.sum(thresh_X) == 4
    assert thresh_X[0][-1] == 0
    assert thresh_X[0][1] == 1


def test_threshold_network_measures_density():

    nm =\
        ThresholdNetworkMeasures(threshold=.2,
                                 threshold_type='abs',
                                 threshold_method='density')

    X = np.array([[.1, .2, .3, -.4],
                  [.2, 0, -.5, 0],
                  [.3, -.5, 0, 0],
                  [-.4, 0, 0, 0]])

    thresh_X = nm._apply_threshold(X)
    assert np.sum(thresh_X) == 2



from ..random_parcellation import RandomParcellation
import numpy as np
import pytest
from ..loaders import load_surf


def test_base():

    fake_geo = [[1, 2],
                [2, 3],
                [3, 4],
                [4, 0],
                [0, 1]]

    parc = RandomParcellation(geo=fake_geo, n_parcels=2,
                              medial_wall_inds=None,
                              medial_wall_mask=None,
                              random_state=None)

    p = parc.get_parc()

    assert len(p) == 5
    assert len(np.unique(p) == 2)


def test_random_seed():

    fake_geo = [[1, 2],
                [2, 3],
                [3, 4],
                [4, 0],
                [0, 1]]

    p = RandomParcellation(geo=fake_geo, n_parcels=2,
                           random_state=1).get_parc()

    for _ in range(5):
        p2 = RandomParcellation(geo=fake_geo, n_parcels=2,
                                random_state=1).get_parc()

        assert np.array_equal(p, p2)


def test_mask_inds():

    fake_geo = [[1, 2],
                [2, 3],
                [3, 4],
                [4, 0],
                [0, 1]]

    p = RandomParcellation(geo=fake_geo, n_parcels=2,
                           medial_wall_inds=[0],
                           random_state=1).get_parc()

    assert p[0] == 0
    assert len(np.unique(p) == 3)


def test_mask():

    fake_geo = [[1, 2],
                [2, 3],
                [3, 4],
                [4, 0],
                [0, 1]]

    p = RandomParcellation(geo=fake_geo, n_parcels=2,
                           medial_wall_mask=[1, 0, 0, 0, 0],
                           random_state=1).get_parc()

    assert p[0] == 0
    assert len(np.unique(p) == 3)


def test_with_load_surf():

    fake_geo = [[1, 2],
                [2, 3],
                [3, 4],
                [4, 0],
                [0, 1]]

    p = RandomParcellation(geo=fake_geo, n_parcels=2,
                           medial_wall_mask=[1, 0, 0, 0, 0],
                           random_state=1)

    s = load_surf(p)
    assert s.shape == (5,)


def test_mask_fail():

    fake_geo = [[1, 2],
                [2, 3],
                [3, 4],
                [4, 0],
                [0, 1]]
    with pytest.raises(IndexError):
        RandomParcellation(geo=fake_geo, n_parcels=2,
                           medial_wall_inds=[0, 0, 0, 1, 0],
                           random_state=1).get_parc()



from ..input import (ModelPipeline, Pipeline, Transformer,
                     Model, ParamSearch, FeatSelector)
from ..input_operations import Duplicate
from sklearn.linear_model import LinearRegression
from ...default.helpers import proc_str_input
import pytest


def test_pipeline_check_duplicate():

    t = Transformer('fake', scope=Duplicate(['1', '2']))
    pipe = Pipeline([t, Model('ridge')])

    assert len(pipe.steps) == 3
    assert pipe.steps[0].scope == '1'
    assert pipe.steps[0].obj == 'fake'
    assert pipe.steps[1].scope == '2'
    assert pipe.steps[1].obj == 'fake'


def test_pipeline_proc_input():

    assert proc_str_input('SOmeThing_reGressor') == 'something'

    with pytest.raises(RuntimeError):
        Pipeline(steps=[Model('SOmeThing_reGressor')])


def test_coarse_check_fail():

    with pytest.raises(RuntimeError):
        Model(obj='not real')

    with pytest.raises(RuntimeError):
        FeatSelector(obj='nope')

    with pytest.raises(RuntimeError):
        FeatSelector(obj='still fake')

    with pytest.raises(TypeError):
        FeatSelector()


def test_get_piece_params():

    t = Transformer('fake', extra='something')
    params = t.get_params()
    assert params['extra'] == 'something'

    t.set_params(extra=5)
    params2 = t.get_params()
    assert params2['extra'] == 5

    with pytest.raises(AttributeError):
        t.extra


def test_modelpipeline_check_duplicate():

    t = Transformer('fake', scope=Duplicate(['1', '2']))
    pipe = ModelPipeline(transformers=t, imputers=None,
                         scalers=None, model='ridge')

    steps = pipe._get_steps()
    assert len(steps) == 3
    assert steps[0].scope == '1'
    assert steps[0].obj == 'fake'
    assert steps[1].scope == '2'
    assert steps[1].obj == 'fake'


def test_pipeline_bad_input():

    with pytest.raises(TypeError):
        Pipeline(6)

    with pytest.raises(IndexError):
        Pipeline([])

    with pytest.raises(RuntimeError):
        Pipeline([Transformer('fake')])

    with pytest.raises(RuntimeError):
        Pipeline(['not real', Model('ridge')])

    with pytest.raises(RuntimeError):
        Pipeline([Model('ridge')], param_search='bad param str')


def test_pipeline_with_sklearn_native_input():
    '''Test some naming behavior w/ saved class attribute'''

    pipe = Pipeline(steps=[Transformer('fake'),
                    ('my_model', LinearRegression())])
    assert pipe.steps[1]._get_step()[0] == 'my_model'

    pipe2 = Pipeline(steps=[Transformer('fake'),
                     ('my_model', LinearRegression())])
    assert pipe2.steps[1]._get_step()[0] == 'my_model'

    del pipe
    pipe3 = Pipeline(steps=[Transformer('fake'),
                     ('my_model', LinearRegression())])
    assert pipe3.steps[1]._get_step()[0] == 'my_model'


def test_get_pipe():

    random_search = ParamSearch('RandomSearch', n_iter=60)
    u_feat = FeatSelector('univariate selection', params=2)
    svm = Model('svm', params=1)
    svm_search_pipe = Pipeline(steps=[u_feat, svm], param_search=random_search)
    Model(svm_search_pipe)



from ..helpers import get_obj_and_params, all_from_avaliable

from sklearn.ensemble import (BaggingClassifier, BaggingRegressor,
                              AdaBoostRegressor, AdaBoostClassifier)
from ...pipeline.ensemble_wrappers import (BPtStackingRegressor,
                                           BPtVotingRegressor,
                                           BPtStackingClassifier,
                                           BPtVotingClassifier)

AVALIABLE = {
        'binary': {
            'basic ensemble': 'voting classifier',
            None: 'voting classifier',
            'voting': 'voting classifier',
            'bagging': 'bagging classifier',
            'adaboost': 'adaboost classifier',
            'stacking': 'stacking classifier',
        },
        'regression': {
                'basic ensemble': 'voting regressor',
                None: 'voting regressor',
                'voting': 'voting regressor',
                'bagging': 'bagging regressor',
                'stacking': 'stacking regressor',
                'adaboost': 'adaboost regressor',
        },
}


ENSEMBLES = {
    'bagging classifier': (BaggingClassifier, ['default']),
    'bagging regressor': (BaggingRegressor, ['default']),
    'adaboost classifier': (AdaBoostClassifier, ['default']),
    'adaboost regressor': (AdaBoostRegressor, ['default']),
    'stacking regressor': (BPtStackingRegressor,
                           ['default']),
    'stacking classifier': (BPtStackingClassifier,
                            ['default']),
    'voting classifier': (BPtVotingClassifier,
                          ['voting classifier']),
    'voting regressor': (BPtVotingRegressor,
                         ['default']),
}

try:

    from imblearn.ensemble import BalancedBaggingClassifier

    AVALIABLE['binary']['balanced bagging'] = 'balanced bagging classifier'

    ENSEMBLES['balanced bagging classifier'] =\
        (BalancedBaggingClassifier, ['default'])

except ImportError:
    pass

# Should be the same
AVALIABLE['categorical'] = AVALIABLE['binary'].copy()


def get_ensemble_and_params(ensemble_str, extra_params, params,
                            **kwargs):

    if ensemble_str == 'basic ensemble' or ensemble_str is None:
        return None, {}

    ensemble, extra_ensemble_params, ensemble_params =\
        get_obj_and_params(ensemble_str, ENSEMBLES, extra_params, params)

    # Slight tweak here, return tuple ensemble, extra_params
    return (ensemble, extra_ensemble_params), ensemble_params


all_obj_keys = all_from_avaliable(AVALIABLE)



from .ScopeObjs import ScopeObj
from sklearn.utils.metaestimators import if_delegate_has_method


class BPtModel(ScopeObj):

    @property
    def _estimator_type(self):
        return self.estimator._estimator_type

    @property
    def feature_importances_(self):
        if hasattr(self.estimator_, 'feature_importances_'):
            return getattr(self.estimator_, 'feature_importances_')
        return None

    @property
    def coef_(self):
        if hasattr(self.estimator_, 'coef_'):
            return getattr(self.estimator_, 'coef_')
        return None

    @property
    def classes_(self):
        if hasattr(self.estimator_, 'classes_'):
            return getattr(self.estimator_, 'classes_')
        return None

    # Every estimator should have at least predict
    def predict(self, X, *args, **kwargs):
        return self.estimator_.predict(X[:, self.inds_], *args, **kwargs)

    @if_delegate_has_method(delegate='estimator_')
    def predict_proba(self, X, *args, **kwargs):
        return self.estimator_.predict_proba(X[:, self.inds_], *args, **kwargs)

    @if_delegate_has_method(delegate='estimator_')
    def decision_function(self, X, *args, **kwargs):
        return self.estimator_.decision_function(X[:, self.inds_],
                                                 *args, **kwargs)

    @if_delegate_has_method(delegate='estimator_')
    def predict_log_proba(self, X, *args, **kwargs):
        return self.estimator_.predict_log_proba(X[:, self.inds_],
                                                 *args, **kwargs)

    @if_delegate_has_method(delegate='estimator_')
    def score(self, X, *args, **kwargs):
        return self.estimator_.score(X[:, self.inds_], *args, **kwargs)



from sklearn.base import BaseEstimator
from sklearn.feature_selection._base import SelectorMixin
import numpy as np
from ..pipeline.helpers import proc_mapping


class FeatureSelector(SelectorMixin, BaseEstimator):

    _needs_mappings = True

    def __init__(self, mask='sets as random features'):
        ''' Custom BPt feature selector for
        integrating in feature selection with a hyper-parameter search.

        Parameters
        ----------
        mask : 'sets as random features' or 'sets as hyperparameters'
            - 'sets as random features': Use random features.

            - 'sets as hyperparameters': Each feature is set as a
                hyperparameter, such that the parameter search can
                tune if each feature is included or not.
        '''

        self.mask = mask
        self.name = 'selector'

    def _proc_mapping(self, mapping):

        if hasattr(self, 'mapping_'):
            return
        else:
            self.mapping_ = mapping

        if len(mapping) > 0:

            mask_0 = np.where(~self.mask)[0]
            mask_1 = np.where(self.mask)[0]

            mask_0 = proc_mapping(mask_0, mapping)
            mask_1 = proc_mapping(mask_1, mapping)

            if len(mask_0) > 0:
                mx_0 = np.max(mask_0)
            else:
                mx_0 = 0

            if len(mask_1) > 0:
                mx_1 = np.max(mask_1)
            else:
                mx_1 = 0

            max_ind = np.max([mx_0, mx_1])

            self.mask = np.zeros(max_ind+1, dtype='bool')
            self.mask[mask_1] = True

        return

    def fit(self, X, y=None, mapping=None):

        if mapping is None:
            mapping = {}

        self.mask = np.array(self.mask)

        threshold = .5
        while np.sum(self.mask >= threshold) == 0:
            threshold -= .001

        self.mask = self.mask >= threshold
        self._proc_mapping(mapping)

        return self

    def _get_support_mask(self):
        return self.mask



from ..BPtLoader import BPtLoader, CompatArray, BPtListLoader
from ...extensions import Identity
import numpy as np
from .helpers import get_fake_mapping, clean_fake_mapping, IdentityListLoader


def test_BPtLoader():

    mapping = get_fake_mapping(10)

    # Test base behavior
    loader = BPtLoader(estimator=Identity(),
                       inds=[0, 1],
                       file_mapping=mapping,
                       n_jobs=1,
                       fix_n_jobs=False,
                       cache_loc=None)
    assert loader._n_jobs == 1

    X = np.arange(10).reshape((5, 2))
    X_trans = loader.fit_transform(X)

    assert X_trans.shape == (5, 8)
    assert X_trans[0][0] == 0
    assert X_trans[0][1] == 0
    assert X_trans[-1][-1] == 9

    # Test to see if passing limited inds works
    loader = BPtLoader(estimator=Identity(),
                       inds=[0],
                       file_mapping=mapping,
                       n_jobs=1,
                       fix_n_jobs=False,
                       cache_loc=None)
    X_trans = loader.fit_transform(X)

    assert X_trans.shape == (5, 5)
    assert np.all(X_trans[:, -1] == np.array([1, 3, 5, 7, 9]))

    # Test passing mapping
    fit_mapping = {0: 1, 1: 0}
    X_trans = loader.fit_transform(X, mapping=fit_mapping)

    assert X_trans.shape == (5, 5)
    assert np.all(X_trans[:, -1] == np.array([0, 2, 4, 6, 8]))

    # Make sure passed on fit mapping is correct
    assert fit_mapping[0] == 4
    assert fit_mapping[1] == [0, 1, 2, 3]

    # Test with mapping to both 0 and 1
    fit_mapping = {0: [0, 1], 1: None}
    X_trans = loader.fit_transform(X, mapping=fit_mapping)
    assert X_trans.shape == (5, 8)
    assert X_trans[0][0] == 0
    assert X_trans[0][1] == 0
    assert X_trans[-1][-1] == 9

    # Make sure passed on fit mapping is correct
    assert fit_mapping[0] == [0, 1, 2, 3, 4, 5, 6, 7]
    assert fit_mapping[1] is None

    # Clean up
    clean_fake_mapping(10)


def test_compat_array():

    x = np.reshape(np.arange(10), (2, 5)).astype('float')
    c = CompatArray(x)

    assert c.shape == (2, 5)
    assert np.array_equal(c[0], x[:, 0])
    assert np.array_equal(c[-1], x[:, -1])

    assert np.array_equal(np.array([], dtype='float').reshape((2, 0)),
                          c.conv_rest_back(rest_inds=[]))
    assert np.array_equal(x[:, 0].reshape((2, 1)),
                          c.conv_rest_back(rest_inds=[0]))
    assert np.array_equal(x, c.conv_rest_back(rest_inds=[0, 1, 2, 3, 4]))


def testBPtListLoader():

    mapping = get_fake_mapping(6)

    loader = BPtListLoader(estimator=IdentityListLoader(),
                           inds=[0],
                           file_mapping=mapping,
                           n_jobs=1,
                           fix_n_jobs=False,
                           cache_loc=None)

    m = {0: 0, 1: 1}
    m_copy = m.copy()
    X = np.arange(6).reshape((3, 2)).astype('float')
    X_trans = loader.fit_transform(X, mapping=m)

    assert X_trans.shape == (3, 5)
    assert X_trans.dtype == 'float'
    assert np.array_equal(X_trans[:, 0], np.array([0, 2, 4]))
    assert np.array_equal(X_trans[:, 1], np.array([0, 2, 4]))
    assert np.array_equal(X_trans[:, 2], np.array([0, 2, 4]))
    assert np.array_equal(X_trans[:, 3], np.array([0, 2, 4]))

    assert loader.n_trans_feats_ == 4
    assert len(loader.X_trans_inds_) == 1
    assert loader.X_trans_inds_[0] == [0, 1, 2, 3]
    assert loader.mapping_ == m_copy
    assert loader.inds_ == [0]
    assert m == loader.out_mapping_
    assert m[0] == [0, 1, 2, 3]



from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array
from sklearn.utils.validation import FLOAT_DTYPES
import numpy as np


class Winsorizer(TransformerMixin, BaseEstimator):
    '''This Scaler performs winzorization, or clipping by feature.

    Parameters
    ----------
    quantile_range : tuple (q_min, q_max), 0.0 < q_min < q_max < 100.0
        Default: (5.0, 95.0), the lower and upper range in which to clip
        values to.

    copy : boolean, optional, default is True
        Make a copy of the data.
    '''

    def __init__(self, quantile_range=(5, 95), copy=True):
        self.quantile_range = quantile_range
        self.copy = copy

    def fit(self, X, y=None):
        X = check_array(X, accept_sparse='csc', estimator=self,
                        dtype=FLOAT_DTYPES, force_all_finite='allow-nan')

        q_min, q_max = self.quantile_range
        if not 0 <= q_min <= q_max <= 100:
            raise ValueError("Invalid quantile range: %s" %
                             str(self.quantile_range))

        self.data_lb_ = np.nanpercentile(X, q_min, axis=0)
        self.data_ub_ = np.nanpercentile(X, q_max, axis=0)

        return self

    def transform(self, X):
        X = check_array(X, accept_sparse=('csr', 'csc'), copy=self.copy,
                        estimator=self, dtype=FLOAT_DTYPES,
                        force_all_finite='allow-nan')

        X = np.clip(X, self.data_lb_, self.data_ub_)
        return X

    def _more_tags(self):
        return {'allow_nan': True}



from .Params import TransitionChoice
import numpy as np

P = {}

# Scalers
P['base standard'] = {'with_mean': True,
                      'with_std': True}

P['base minmax'] = {'feature_range': (0, 1)}

P['base robust'] = {'quantile_range': (5, 95)}

P['base winsorize'] = {'quantile_range': (1, 99)}

P['robust gs'] =\
        {'quantile_range':
         TransitionChoice([(x, 100-x) for x in np.arange(1, 40)])}

P['winsorize gs'] =\
        {'quantile_range':
         TransitionChoice([(x, 100-x) for x in np.arange(1, 40)])}

P['base yeo'] = {'method': 'yeo-johnson',
                 'standardize': True}

P['base boxcox'] = {'method': 'box-cox',
                    'standardize': True}

P['base quant norm'] = {'output_distribution': 'normal'}

P['base quant uniform'] = {'output_distribution': 'uniform'}



_shared_docs = {}

_shared_docs[
    "problem_spec"
] = """problem_spec : :class:`ProblemSpec` or 'default', optional
        This parameter accepts an instance of the
        params class :class:`ProblemSpec`.
        The ProblemSpec is essentially a wrapper
        around commonly used
        parameters needs to define the context
        the model pipeline should be evaluated in.
        It includes parameters like problem_type, scorer, n_jobs,
        random_state, etc...

        See :class:`ProblemSpec` for more information
        and for how to create an instance of this object.

        If left as 'default', then will initialize a
        ProblemSpec with default params.

        ::

            default = "default"

"""

_shared_docs[
    "problem_spec_params"
] = """problem_spec_params : :class:`ProblemSpec` params, optional
        You may also pass any valid problem spec argument-value pairs here,
        in order to override a value in the passed :class:`ProblemSpec`.
        Overriding params should be passed in kwargs style, for example:

        ::

            func(..., problem_type='binary')

    """



from pandas.core.dtypes.base import ExtensionDtype, register_extension_dtype
import numpy as np


class DataFileDtypeType(type):
    pass


@register_extension_dtype
class DataFileDtype(ExtensionDtype):

    kind = 'O'
    str = "|O08"
    base = np.dtype("O")
    subdtype = None
    type = DataFileDtypeType
    name = 'file'

    isbuiltin = 0
    isnative = 0

    _metadata = ()
    na_value = np.nan

    def __str__(self):
        """
        Return a string representation for a particular Object
        """
        return self.name

    def __repr__(self):
        """
        Return a string representation for a particular object.
        """
        return self.name

    def __getstate__(self):
        return {k: getattr(self, k, None) for k in self._metadata}

    @classmethod
    def construct_array_type(cls):
        """
        Return the array type associated with this dtype.
        Returns
        -------
        type
        """

        from .data_file_array import DataFileArray
        return DataFileArray

    @classmethod
    def construct_from_string(cls, string):
        if not isinstance(string, str):
            raise TypeError(
                f"'construct_from_string' expects a string, got {type(string)}"
            )

        if string != cls.name:
            raise TypeError(
                f"Cannot construct a 'DataFileDType' from '{string}'")

        return cls()

    def __hash__(self) -> int:
        return hash(str(self))

    def __setstate__(self, state) -> None:
        # for pickle compat. __get_state__ is defined in the
        # PandasExtensionDtype superclass and uses the public properties to
        # pickle -> need to set the settable private ones here (see GH26067)
        # self._tz = state["tz"]
        # self._unit = state["unit"]

        pass

    @classmethod
    def _from_values_or_dtype(cls, values=None, categories=None,
                              ordered=None, dtype=None):
        print('_from_values_or_dtype')

    def _get_common_dtype(self, dtypes):
        return self

    @property
    def _is_boolean(self) -> bool:
        return False



from ..util import BPtInputMixIn
from copy import copy


class Select(list, BPtInputMixIn):
    '''The Select object is an BPt specific Input Wrapper designed
    to allow hyper-parameter searches to include not
    just choice of hyper-parameter,
    but also choosing between objects
    (as well as their relevant distributions).

    Select is used to cast lists of base :class:`ModelPipeline`
    or :class:`Pipeline` pieces as different options.
    Consider a simple example, for specifying a selection
    between two different :class:`Models<Model>`

    ::

        model = Select([Model('linear'), Model('random forest')])

    In this example, the model passed to :class:`Pipeline`
    or :class:`Model_Pipeline`
    becomes a meta object
    for selecting between the two base models.
    Note: this does require a :class:`ParamSearch`
    object be passed to :class:`Pipeline`.
    Notably as well, if further param distributions
    are defined within say the `Model('random forest')`,
    those will still be optimized, allowing for
    potentially even a hyper-parameter search to
    select hyper-parameter distribution...
    (i.e., if both select options had the same base model obj,
    but only differed in the passed hyper-param distributions) if one were
    so inclined...

    Other notable features of Select are, you are not limited to
    passing only two options,
    you can pass an arbitrary number... you can even,
    and I'm not even sure I want to tell you this...
    pass nested calls to Select... i.e., one of the
    Select options could be another Select, with say
    another Select...

    Lastly, explicitly note that Select is not restricted for use with Models,
    it can be used on any of
    the base class:`Pipeline` piece params
    (i.e., every param but param_search and cache...).

    Notes
    -------
    This object cannot be used to wrap custom objects.

    '''

    def __repr__(self):
        return 'Select(' + super().__repr__() + ')'

    def __str__(self):
        return self.__repr__()

    def _check_args(self):

        for option in self:
            option._check_args()

        # Make sure all constructors the same
        if len(set([option._constructor for option in self])) != 1:
            raise RuntimeError('Select must be composed '
                               'of the same type of pieces!')

    def _uniquify(self):

        for i, option in enumerate(self):

            # Recursive check first
            if hasattr(option, '_uniquify'):
                option._uniquify()

            # Then replace with copy
            self[i] = copy(option)


    @property
    def _constructor(self):

        # We know all constructors are the same from _check_args
        return self[0]._constructor


class Duplicate(list, BPtInputMixIn):
    '''The Duplicate object is an BPt specific Input wrapper.
    It is designed to be cast on a list of valid scope parameters, e.g.,

    ::

        scope = Duplicate(['float', 'cat'])

    Such that the corresponding pipeline piece will be duplicated for every
    entry within Duplicate. In this case, two copies of the base object will be
    made, where both have the same remaining non-scope params
    (i.e., obj, params, extra_params),
    but one will have a scope of 'float' and the other 'cat'.

    Consider the following extended example, where loaders is being specified
    when creating an instance of :class:`Pipeline`:

    ::

        loaders = Loader(obj='identity', scope=Duplicate(['float', 'cat']))

    Is transformed in post processing / equivalent to

    ::

        loaders = [Loader(obj='identity', scope='float'),
                   Loader(obj='identity', scope='cat')]

    '''

    def __repr__(self):
        return 'Duplicate(' + super().__repr__() + ')'

    def __str__(self):
        return self.__repr__()


class Pipe(list, BPtInputMixIn):
    '''The Pipe object is an BPt specific Input wrapper, designed
    for now to work specifically within :class:`Loader`.
    Because loader
    objects within BPt are designed to work on single files at a time,
    and further are restricted in that they must go
    directly from some arbitrary
    file, shape and charteristics to outputted
    as a valid 2D (# Subjects X # Features) array,
    it restricts potential sequential compositions.
    Pipe offers some utility towards
    building sequential compositions.

    For example, say one had saved 4D neuroimaging fMRI timeseries,
    and they wanted
    to first employ a loader to extract timeseries by ROI
    (with say hyper-parameters defined to select which ROI to use),
    but then wanted to use
    another loader to convert the timeseries ROI's to a correlation matrix,
    and only then pass
    along the output as 1D features per subject.
    In this case, the Pipe wrapper is a great candidate!

    Specifically, the pipe wrapper works at the level of defining a
    specific Loader, where basically
    you are requesting that the loader you want to use be a
    Pipeline of a few different loader options,
    where the loader options are ones compatible in
    passing input to each other, e.g., the output from
    fit_transform as called on the ROI extractor is valid input
    to fit_transform of the Timeseries creator,
    and lastly the output from fit_transform of the
    Timeseries creator valid 1D feature array per subjects output.

    Consider the example in code below, where we
    assume that 'rois' is the ROI extractor,
    and 'timeseries' is the correlation matrix
    creator object (where these could be can valid loader str, or
    custom user passed objects)

    ::

        loader = Loader(obj = Pipe(['rois', 'timeseries']))

    We only passed arguments for obj above, but in our toy example
    as initially described we wanted to
    further define parameters for a parameter search across both objects.
    See below for what different options
    for passing corresponding parameter distributions are:

    ::

        # Options loader1 and loader2 tell it explicitly no params

        # Special case, if just default params = 0, will convert to 2nd case
        loader1 = Loader(obj = Pipe(['rois', 'timeseries']),
                         params = 0)

        # You can specify just a matching list
        loader2 = Loader(obj = Pipe(['rois', 'timeseries']),
                         params = [0, 0])

        # Option 3 assumes that there are pre-defined valid class param dists
        # for each of the base objects
        loader3 = Loader(obj = Pipe(['rois', 'timeseries']),
                         params = [1, 1])

        # Option 4 lets set params for the 'rois' object, w/ custom param dists
        loader4 = Loader(obj = Pipe(['rois', 'timeseries']),
                         params = [{'some custom param dist'}, 0])

    Note that still only one scope may be passed, and that scope will
    define the scope of the new combined loader.
    Also note that if extra_params is passed, the same extra_params will
    be passed when creating both individual objects.
    Where extra params behavior is to add its contents, only when the
    name of that param appears in the base classes init, s.t.
    there could exist a case where, if both 'rois' and 'timeseries'
    base objects had a parameter with the same name, passing a
    value for that name in extra params would update them
    both with the passed value.
    '''

    def __repr__(self):
        return 'Pipe(' + super().__repr__() + ')'

    def __str__(self):
        return self.__repr__()


class ValueSubset(BPtInputMixIn):
    ''' ValueSubset is special wrapper class for BPt designed to work with
    :ref:`Subjects` style input. It is used to select a subset of subjects
    based on a loaded column's value.

    Parameters
    -----------
    name : str
        The name of the loaded column in :class:`Dataset`
        in which the values refer to.

        In the case of selecting values from
        an overlap of multiple columns, you can
        first create a new overlap column.
        See method :func:`Dataset.add_unique_overlap`,
        then pass the name of the overlap column here.

    values : str, float or list of
        This parameter refers to either a single or
        list of values in which if the loaded
        subject has the value, or one of the values,
        in the column selected
        via name, then that subject will be selected.

    decode_values : bool, optional
        This parameter is boolean flag which represents if encoded
        values should be
        used or not. What this asks is that should the actual
        ordinal post encoded
        value be specified, or the should the value be
        set based on the original
        encoded name.

        This parameter is only relevant assuming the underlying
        name column was encoding using an encoding method
        from :class:`Dataset`, for example :func:`Dataset.to_binary`
        or :func:`Dataset.ordinalize`.

        ::

            default = False

    Examples
    ---------

    This wrapper can be used as follows, just specify an object as

     ::

        ValueSubset(name, values)

    Where name is the name of a loaded non input categorical column / feature,
    and value is the subset of values from that column to select subjects by.
    E.g., if you wanted to select just subjects of a specific sex,
    and assuming the variable was properly loaded,

    ::

        subjects = ValueSubset('sex', 0)

    Which would specify only subjects with 'sex' equal to 0.
    You may also optionally pass more than one value to values, E.g.,

    ::

        subjects = Values_Subset(name='site', values=[0, 1, 5])

    Would select the subset of subjects from sites 0, 1 and 5.

    We can also consider using the decode_values parameter.
    For example, let's say sex originally had values 'M' and 'F'
    and then :func:`Dataset.binarize` was used to
    transform the values internally to 0 and 1.
    If decode_values is set
    as the default value of False,
    then you must pass value = 0 or 1, but
    if decode_values = True,
    then it allows passing value = 'M' or 'F'. E.g.,

    ::

        subjects = ValueSubset('sex', 'M', decode_values=True)

    '''

    def __init__(self, name, values, decode_values=False):

        self.name = name
        self.values = values
        self.decode_values = decode_values

        if isinstance(self.name, list):
            raise ValueError('name cannot be list / array-like!')

        if not isinstance(self.decode_values, bool):
            raise ValueError('decode_values must be a bool')

    def __repr__(self):
        return 'ValueSubset(name=' + str(self.name) + ', values=' + \
          str(self.values) + ', decode_values=' + \
          str(self.decode_values) + ')'

    def __str__(self):
        return self.__repr__()


class Intersection(list, BPtInputMixIn):
    '''The Intersection class is a special
    wrapper class used to request the intersection
    of two valid arguments for :ref:`Subjects`.

    For example:

    ::

        subjects = Intersection(['train',
                                [1, 2, 3]])

    Would specify subjects as the intersection of the loaded
    train subjects and subjects 1, 2 and 3.
    '''

    def __repr__(self):
        return 'Intersection(' + super().__repr__() + ')'

    def __str__(self):
        return self.__repr__()



import pandas as pd
from pandas.util._decorators import doc
from .Dataset import _shared_docs
from .helpers import (base_load_subjects)
from ..util import conv_to_list
from ..main.input_operations import Intersection, ValueSubset
from itertools import product


def _apply_only_level(self, subjects, only_level):

    # If only level not None and MultiIndex - only keep
    if isinstance(self.index, pd.MultiIndex) and only_level is not None:
        drop_names = list(self.index.names)
        drop_names.remove(self.index.names[only_level])
        subjects = subjects.droplevel(drop_names)

    return set(list(subjects))


def _get_nan_loaded_subjects(self, only_level):

    # Get nan subjects
    nan_subjects = self[pd.isnull(self[:]).any(axis=1)].index

    # Apply only level + convert to sat
    nan_subjects = self._apply_only_level(nan_subjects, only_level)

    return nan_subjects


def _get_value_subset_loaded_subjects(self, subjects, only_level):

    if subjects.name not in list(self):
        raise KeyError('Passed ValueSubset name: ' +
                       repr(subjects.name) +
                       ' is not loaded')

    # Get the relevant series
    data = self.get_values(subjects.name, dropna=False,
                           decode_values=subjects.decode_values)

    # Extract the values as list
    values = conv_to_list(subjects.values)

    # Get subjects by values
    loaded_subjects = data[data.isin(values)].index

    # Apply only level + convert to set
    loaded_subjects =\
        self._apply_only_level(loaded_subjects, only_level)

    return loaded_subjects


def _get_base_loaded_subjects(self, subjects, only_level):

    loaded_subjects = base_load_subjects(subjects)

    # In multi-index case
    if isinstance(self.index, pd.MultiIndex):

        # If want multi-index returned
        if only_level is None:
            loaded_subjects = set(self.loc[loaded_subjects].index)

    # Non multi-index case try to cast to correct type
    else:

        dtype = self.index.dtype.name
        if 'int' in dtype:
            loaded_subjects = {int(s) for s in loaded_subjects}
        elif 'float' in dtype:
            loaded_subjects = {float(s) for s in loaded_subjects}
        elif dtype == 'object':
            loaded_subjects = {str(s) for s in loaded_subjects}
        else:
            raise RuntimeError('Index data type:' + dtype + ' '
                               'is not currently supported!')

    return loaded_subjects


def _return_subjects_as(self, subjects, return_as, only_level):

    if return_as == 'set':
        return subjects

    # If not set, treat as pandas index or flat index case
    subjects = sorted(list(subjects))

    # If not multi index, just cast and return
    if not isinstance(self.index, pd.MultiIndex):
        return pd.Index(data=subjects, name=self.index.name)

    # Multi index case, with only level
    if only_level is not None:
        return pd.Index(data=subjects,
                        name=self.index.names[only_level])

    # RConvert to multi index
    subjects = list(map(tuple, subjects))
    multi_index = pd.MultiIndex.from_tuples(
        subjects, names=self.index.names)

    if return_as == 'flat index':
        return multi_index.to_flat_index()
    else:
        return multi_index


@doc(subjects=_shared_docs['subjects'])
def get_subjects(self, subjects, return_as='set', only_level=None):
    '''Method to get a set of subjects, from
    a set of already loaded ones, or from a saved location.

    Parameters
    -----------
    {subjects}

    return_as : ['set', 'index', 'flat index'], optional

        - 'set'
            Return as set of subjects.

        - 'index'
            Return as sorted :class:`pandas.Index`,
            or if underlying multi-index as a
            :class:`pandas.MultiIndex`.

        - 'flat index'
            Return as sorted :class:`pandas.Index`,
            note that if not an underlying multi-index
            this will return the same result as
            'index', when MultiIndex, will return the index as a flat
            Index of tuples.

    only_level : int or None, optional
        This parameter is only relevant when the
        underlying index is a MultiIndex.

        Note: this param is not relevant
        when using special tuple style input for subjects.

        ::

            default = None

    Returns
    ----------
    subjects : set, :class:`pandas.Index` or :class:`pandas.MultiIndex`
        Based on value of return_as, returns as
        a set of subjects, sorted :class:`pandas.Index`,
        sorted :class:`pandas.MultiIndex` or flattened and
        sorted :class:`pandas.Index` representing a
        :class:`pandas.MultiIndex`.

    '''

    # input validation
    if return_as not in ['set', 'index', 'flat index']:
        raise TypeError('Invalid parameter passed to return as!')

    # Check for passed intersection case
    if isinstance(subjects, Intersection):

        subjects_list =\
            [self.get_subjects(s, return_as='set', only_level=only_level)
                for s in subjects]

        loaded_subjects = set.intersection(*subjects_list)

        # Return as requested
        return self._return_subjects_as(loaded_subjects,
                                        return_as=return_as,
                                        only_level=only_level)

    if isinstance(subjects, tuple):

        if len(subjects) != len(self.index.names):
            raise RuntimeError('Passed special tuple must match the '
                               'number of MultiIndex levels.')

        # Proc each ind seperately
        inds = []
        for i, subject_arg in enumerate(subjects):
            inds.append(self.get_subjects(subject_arg,
                                          return_as='set',
                                          only_level=i))
        # Create set of sets
        loaded_subjects = set(product(*inds))

        # Return as requested, note only_level fixed as None
        return self._return_subjects_as(loaded_subjects,
                                        return_as=return_as,
                                        only_level=None)

    # Compute all subjects at correct only level
    all_subjects = self._apply_only_level(self.index, only_level)

    # Check if None
    if subjects is None:
        loaded_subjects = set()

    # Check for special keywords
    elif isinstance(subjects, str) and subjects == 'nan':
        loaded_subjects =\
            self._get_nan_loaded_subjects(only_level=only_level)

    elif isinstance(subjects, str) and subjects in ['all', 'default']:
        loaded_subjects = all_subjects

    elif isinstance(subjects, str) and subjects == 'train':
        if not hasattr(self, 'train_subjects') or self.train_subjects is None:
            raise RuntimeError('Train subjects undefined')
        loaded_subjects =\
            self._apply_only_level(self.train_subjects, only_level)

    elif isinstance(subjects, str) and subjects == 'test':
        if not hasattr(self, 'test_subjects') or self.test_subjects is None:
            raise RuntimeError('Test subjects undefined')
        loaded_subjects =\
            self._apply_only_level(self.test_subjects, only_level)

    # Check for Value Subset or Values Subset
    elif isinstance(subjects, ValueSubset):
        loaded_subjects =\
            self._get_value_subset_loaded_subjects(subjects,
                                                   only_level=only_level)
    else:
        loaded_subjects =\
            self._get_base_loaded_subjects(subjects, only_level=only_level)

    # loaded_subjects are a set here, though still need to only
    # get the overlap of actually loaded subjects here
    loaded_subjects.intersection_update(all_subjects)

    # Return based on return as value
    return self._return_subjects_as(loaded_subjects, return_as,
                                    only_level=only_level)




from copy import deepcopy
from .model_params import P as model_params
from .ensemble_params import P as ensemble_params
from .feat_selector_params import P as feat_selector_params
from .imputer_params import P as imputer_params
from .scaler_params import P as scaler_params
from .transformer_params import P as transformer_params

# Extensions
from ...extensions.mv_transformer.mv_transformer_params import P as mv_params

PARAMS = {}
PARAMS.update(model_params)
PARAMS.update(ensemble_params)
PARAMS.update(feat_selector_params)
PARAMS.update(scaler_params)
PARAMS.update(imputer_params)
PARAMS.update(transformer_params)
PARAMS.update(mv_params)


def get_base_params(str_indicator):

    base_params = deepcopy(PARAMS[str_indicator])
    return base_params


def proc_params(base_params, prepend):

    if isinstance(base_params, int):
        return {}

    params = {prepend + '__' + key: base_params[key] for key in
              base_params}

    return params



from .dataset.Dataset import Dataset, read_csv
from .main.input import (Loader, Imputer, Scaler,
                         Transformer,
                         FeatSelector, Model, Ensemble,
                         Param_Search, ParamSearch, ModelPipeline,
                         Model_Pipeline, Pipeline,
                         Problem_Spec, ProblemSpec,
                         CV, CVStrategy, Feat_Selector)
from .main.funcs import (get_estimator, cross_validate,
                         cross_val_score, evaluate)

from .main.input_operations import (Select, Duplicate, Pipe, ValueSubset,
                                    Intersection)
from .main.compare import Compare, Option, CompareDict
from .main.BPtEvaluator import BPtEvaluator
from . import p

from pandas import read_pickle

__author__ = "sahahn"
__version__ = "2.0.0"
__all__ = ["Dataset", "Loader",
           "Imputer", "Scaler", "Transformer", 'get_estimator',
           "FeatSelector", "Model",
           "Ensemble", "Param_Search", "ParamSearch",
           "Model_Pipeline", "Problem_Spec", "ProblemSpec", "Select",
           "Duplicate", "Pipe", "ValueSubset",
           "CV", "CVStrategy", "Intersection",
           "cross_validate", "cross_val_score", "evaluate",
           'p', "Feat_Selector", 'ModelPipeline', 'Pipeline',
           'BPtEvaluator', 'read_pickle', 'Compare', 'Option',
           'CompareDict', 'read_csv']



import numpy as np
from ..BPtFeatureSelector import BPtFeatureSelector
from .helpers import FakeSelector


def test_bptfeature_selector():

    estimator = FakeSelector(mask=np.array([True, False, True]))
    fs = BPtFeatureSelector(estimator=estimator, inds=[0, 1, 2],
                            cache_loc=None)

    # Test base behavior
    X = np.ones((3, 3))
    X[:, 0] = 0
    X[:, 2] = 2

    X_trans = fs.fit_transform(X)
    assert X_trans.shape == (3, 2)
    assert X_trans[0][0] == 0
    assert X_trans[0][1] == 2

    # Try with scope
    X = np.ones((3, 4))
    X[:, 0] = 0
    X[:, 2] = 2
    X[:, 3] = 3

    X_trans = fs.fit_transform(X)
    fs = BPtFeatureSelector(estimator=estimator, inds=[0, 1, 3],
                            cache_loc=None)

    X_trans = fs.fit_transform(X)
    assert X_trans.shape == (3, 3)
    assert X_trans[0][0] == 0
    assert X_trans[0][1] == 3
    assert X_trans[0][2] == 2

    # Try adding fit mapping, just flip 0 and 2
    fit_mapping = {0: 2, 1: 1, 2: 0, 3: 3}
    X_trans = fs.fit_transform(X, mapping=fit_mapping)
    assert X_trans.shape == (3, 3)
    assert X_trans[0][0] == 1
    assert X_trans[0][1] == 3
    assert X_trans[0][2] == 0

    X_rev = fs.inverse_transform(X_trans)
    assert X_rev[0][0] == 0
    assert X_rev[0][1] == 1
    assert X_rev[0][2] == 0
    assert X_rev[0][3] == 3



from ..from_sklearn import load_boston
from ...dataset.Dataset import Dataset


def test_load_boston():

    data = load_boston()

    assert data.shape == (506, 14)
    assert data.train_subjects is not None
    assert isinstance(data, Dataset)



from BPt.pipeline.BPtFeatureSelector import BPtFeatureSelector
from ..main.input_operations import Pipe, Select

from .helpers import (param_len_check, replace_model_name)
from ..default.helpers import process_params_by_type, proc_type_dep_str
from ..util import conv_to_list

from .ScopeObjs import ScopeTransformer
from .BPtModel import BPtModel
from sklearn.ensemble import VotingClassifier, VotingRegressor

from sklearn.pipeline import Pipeline
from copy import deepcopy

from .Selector import selector_wrapper
from sklearn.compose import TransformedTargetRegressor
from .BPtSearchCV import wrap_param_search
from .ensemble_wrappers import EnsembleWrapper
from collections import defaultdict

import numpy as np


def check_for_duplicate_names(objs_and_params):
    '''Checks for duplicate names within an objs_and_params type obj'''

    names = [c[0] for c in objs_and_params]

    # If any repeats
    if len(names) != len(set(names)):
        new_objs_and_params = []

        for obj in objs_and_params:
            name = obj[0]

            if name in names:

                cnt = 0
                used = [c[0] for c in new_objs_and_params]
                while name + str(cnt) in used:
                    cnt += 1

                # Need to change name within params also
                base_obj = obj[1][0]
                base_obj_params = obj[1][1]

                new_obj_params = {}
                for param_name in base_obj_params:

                    p_split = param_name.split('__')
                    new_param_name = p_split[0] + str(cnt)
                    new_param_name += '__' + '__'.join(p_split[1:])

                    new_obj_params[new_param_name] =\
                        base_obj_params[param_name]

                new_objs_and_params.append((name + str(cnt),
                                           (base_obj, new_obj_params)))

            else:
                new_objs_and_params.append(obj)

        return new_objs_and_params
    return objs_and_params


def process_input_types(obj_strs, param_strs, scopes):

    # I think want to fill in missing param_strs w/ 0's here
    param_strs = param_len_check(obj_strs, param_strs)
    return obj_strs, param_strs, scopes


def get_scope_name(scope):

    scope_name = ' ' + str(scope)
    if scope_name == ' all':
        scope_name = ''
    elif len(scope_name) > 10:
        scope_name = scope_name[:10] + '...'

    return scope_name


def add_estimator_to_params(passed_params):
    '''Process the params, all have the same base name, but now need estimator
    to correspond to changing the correct params, e.g. if
    my_loader__some_param becomes my_loader__estimator__some_param
    '''

    params = {}
    for key in passed_params:

        split_key = key.split('__')
        new_split_key = [split_key[0], 'estimator'] + split_key[1:]
        new_key = '__'.join(new_split_key)

        # Add to new under new name
        params[new_key] = passed_params[key]

    return params


class Constructor():

    def __init__(self, user_passed_objs, dataset, spec):
        # problem_type, random_state, n_jobs are stored in spec

        # Class values
        self.user_passed_objs = user_passed_objs
        self.dataset = dataset
        self.spec = spec.copy()

        # This value with be replaced in child classes
        self.AVAILABLE = None

    def get_inds(self, scope):

        # @ TODO ADD warning when a scope returns None here?
        return self.dataset._get_data_inds(ps_scope=self.spec['scope'],
                                           scope=scope)

    def process(self, params):
        '''params is a list of Param piece classes, or potentially a list
        like custom object,
        e.g., Select. Each actual param object might vary a bit on
        what it includes'''

        if params is None or (isinstance(params, list) and len(params) == 0):
            return [], {}

        # In case of recursive calls, make sure params are as list
        params = conv_to_list(params)

        # Check for select
        select_mask = np.array([isinstance(param, Select) for param in params])
        if select_mask.any():
            return self._process_with_select(params, select_mask)

        # Otherwise, process as normal
        return self._process(params)

    def _process_with_select(self, params, select_mask):

        # Check & Process for Select()
        # All params have been conv_to_list, s.t. if Select is present
        # Then it is as [Select([...])] or [Select([...]), ...]

        # Init objs as empty list of correct size w/ just place holders
        objs = [None for i in range(len(params))]

        # Process everything but the select groups first
        # Putting the recursive call to process here... even though I
        # think call to
        # just _process could work too. The logic is, I don't think it will
        # hurt,
        # and if more options beyond Select are added to process later... this
        # will hopefully cover those cases.
        non_select_objs, obj_params =\
            self.process([i for idx, i in enumerate(params) if
                         not select_mask[idx]])

        # Update right spot in objs
        for obj, ind in zip(non_select_objs, np.where(~select_mask)[0]):
            objs[ind] = obj

        # Split up select params
        select_params = [i for idx, i in enumerate(params) if select_mask[idx]]

        # Next, process each group of select params seperately
        cnt = 0
        for s_params, ind in zip(select_params, np.where(select_mask)[0]):

            # Recursive call to process... s.t., can handle nested Select...
            # oh god
            s_objs, s_obj_params = self.process(list(s_params))

            # Wrap in selector object
            name = self.name + '_selector' + str(cnt)
            s_obj, s_obj_params = selector_wrapper(s_objs, s_obj_params, name)

            # Can update final params
            objs[ind] = s_obj
            obj_params.update(s_obj_params)
            cnt += 1

        return objs, obj_params

    # Overide in children class
    def _process(self, params):
        return [], {}

    def _get_obj_param(self, get_func, param):

        # Unpack
        name, param_str = param.obj, param.params
        extra_params = param.extra_params

        # Custom case
        if 'Custom ' in name:
            return self._get_user_passed_obj_params(name, param_str,
                                                    extra_params)

        # Get original number of feat keys
        # just used for special feature selector case right now.
        num_feat_keys = 0
        if hasattr(param, 'scope'):
            num_feat_keys = len(self.get_inds(param.scope))

        # Get obj from func, where any un-needed params
        # are just not used.
        obj = get_func(name, extra_params, param_str,
                       random_state=self.spec['random_state'],
                       num_feat_keys=num_feat_keys)

        return (name, obj)

    def _get_objs_and_params(self, get_func, params):
        '''Helper function to grab scaler / feat_selectors and
        their relevant parameter grids'''

        # Make the object + params based on passed settings
        objs_and_params = []
        for param in params:
            objs_and_params.append(self._get_obj_param(get_func, param))

        # Perform extra proc, to split into objs and merged param dict
        objs, params = self._proc_objs_and_params(objs_and_params)

        return objs, params

    def _get_user_passed_obj_params(self, name, param, extra_params):

        # Get copy of user_obj
        user_obj = deepcopy(self.user_passed_objs[name])

        # Proc as necc. user passed params
        extra_user_obj_params, user_obj_params =\
            process_params_by_type(obj_str=name,
                                   base_params=deepcopy(param),
                                   extra_params=extra_params)

        # If passing a user object, kind of stupid to pass default, non search
        # params, via a dict..., but hey give it a try
        try:
            user_obj.set_params(**extra_user_obj_params)
        except AttributeError:
            pass

        # Return in obj_param format
        return (name, (user_obj, user_obj_params))

    def _proc_objs_and_params(self, objs_and_params):

        # If two of same object passed, change name
        objs_and_params = check_for_duplicate_names(objs_and_params)

        # Construct the obj as list of (name, obj) tuples, checking each base
        # objects params
        objs = [(c[0], self._check_params(c[1][0])) for c in objs_and_params]

        # Grab the params, and merge them into one dict of all params
        params = {k: v for params in objs_and_params
                  for k, v in params[1][1].items()}

        return objs, params

    def _check_params(self, obj):

        if hasattr(obj, 'random_state'):
            setattr(obj, 'random_state', self.spec['random_state'])

        if hasattr(obj, 'n_jobs'):
            setattr(obj, 'n_jobs', self.spec['n_jobs'])

        if hasattr(obj, '_needs_cat_inds'):
            if getattr(obj, '_needs_cat_inds'):
                obj.set_params(cat_inds=self.get_inds('category'))

        return obj

    def replace_base_estimator(self, objs, obj_params, params):

        for i in range(len(objs)):

            if (hasattr(objs[i][1], 'estimator') and
               params[i].base_model is not None):

                # Proc base model type
                model_spec = self.spec.copy()
                base_model_type = params[i].base_model_type
                if base_model_type is not None:
                    if base_model_type == 'default':

                        # Default regression
                        model_spec['problem_type'] = 'regression'

                        # Only change to categorical
                        # if all cols within scope are categorical
                        all_cat = self.dataset._is_data_cat(
                            ps_scope=self.spec['scope'],
                            scope=params[i].scope)

                        if all_cat:
                            model_spec['problem_type'] = 'categorical'

                    else:
                        model_spec['problem_type'] = base_model_type

                # Grab the base estimator
                base_model_obj = ModelConstructor(self.user_passed_objs,
                                                  self.dataset,
                                                  model_spec)
                base_objs, base_params =\
                    base_model_obj.process(params[i].base_model)

                # Replace the estimator
                objs[i][1].estimator = base_objs[0][1]

                # Replace the model name in the params with estimator
                base_params = replace_model_name(base_params)

                # Add the name of the object pre-pended
                base_params = {objs[i][0] + '__' + p_name: base_params[p_name]
                               for p_name in base_params}

                obj_params.update(base_params)

        return objs, obj_params

    def _check_scope_all(self, scope):
        '''If scope is all or functionally all,
        return Ellipsis. If not, will return
        the actual scope inds.'''

        # If keyword all, return True
        if scope == 'all':
            return Ellipsis

        # Otherwise check for functionally
        # all case.
        inds = self.get_inds(scope)
        all_inds = self.get_inds('all')

        # Since inds unique, if same length, then same
        if len(inds) == len(all_inds):
            return Ellipsis

        return inds

    def _make_col_version(self, objs, params, input_params, Wrapper):
        '''@TODO re-use pieces from this and checking for a Scope model, comb
        simmilar functionality into one method.'''

        # Make objs + params
        col_objs, col_params = [], {}

        for i in range(len(objs)):

            # Unpack name and obj
            name, obj = objs[i][0], objs[i][1]

            # Make sure n_jobs and random_state are set in the base object
            if hasattr(obj, 'n_jobs'):
                setattr(obj, 'n_jobs', self.spec['n_jobs'])
            if hasattr(obj, 'random_state'):
                setattr(obj, 'random_state', self.spec['random_state'])

            if hasattr(input_params[i], 'scope'):

                # Proc scope
                scope = input_params[i].scope

                # Sets to either inds or Ellipsis if all
                inds = self._check_scope_all(scope)

                # Get scope name
                scope_name = get_scope_name(scope)

            elif hasattr(input_params[i], 'scopes'):

                # Proc scope
                scopes = input_params[i].scopes

                # Sets to either inds or Ellipsis if all
                inds = [self._check_scope_all(scope) for scope in scopes]

                # Get scope name
                scope_name = ''

            else:
                raise RuntimeError('No scope or scopes in object')

            if hasattr(input_params[i], 'cache_loc'):
                cache_loc = getattr(input_params[i], 'cache_loc')
            else:
                cache_loc = None

            # Create the col obj as a the passed Wrapper
            col_obj = (name + scope_name,
                       Wrapper(estimator=obj,
                               inds=inds,
                               cache_loc=cache_loc))
            col_objs.append(col_obj)

            # Change any associated params
            for key in params:
                split_key = key.split('__')

                # If this is an associated param w/ this obj
                if split_key[0] == name:

                    # Replace name with estimator
                    name = split_key[0] + scope_name
                    split_key[0] = 'estimator'

                    # Create new name
                    new_name = '__'.join([name] + split_key)

                    # Save under new name
                    col_params[new_name] = params[key]

        return col_objs, col_params


class TypeConstructor(Constructor):

    def _base_type_process(self, params, func):

        # Check / replace by type
        params = self._check_type(params)

        # Then call get objs and params
        objs, obj_params =\
            self._get_objs_and_params(func, params)

        return objs, obj_params

    def _check_type(self, params):

        for p in range(len(params)):
            params[p].obj =\
                proc_type_dep_str(params[p].obj, self.AVAILABLE,
                                  self.spec['problem_type'])
        return params


class ModelConstructor(TypeConstructor):

    name = 'models'

    def _process(self, params):

        from ..default.options.models import (get_base_model_and_params,
                                              AVALIABLE)
        self.AVAILABLE = AVALIABLE.copy()

        # If any ensembles
        ensemble_mask = np.array([hasattr(params[i], 'models')
                                  for i in range(len(params))])

        # Separate non-ensemble objs and obj_params
        non_ensemble_params = [i for idx, i in enumerate(params)
                               if not ensemble_mask[idx]]

        # Run initial base type process on the non ensemble models
        non_ensemble_objs, non_ensemble_obj_params =\
            self._base_type_process(non_ensemble_params,
                                    get_base_model_and_params)

        # Perform wrap checks on non_ensembles
        non_ensemble_objs, non_ensemble_obj_params =\
            self.check_wraps(non_ensemble_objs,
                             non_ensemble_obj_params,
                             non_ensemble_params)

        # If just models, i.e., no ensembles to process,
        # can return here just the non_ensemble_objs + params
        if not ensemble_mask.any():
            return non_ensemble_objs, non_ensemble_obj_params

        # Assume ensemble case from here
        # Process the ensemble objs and params
        ensembled_objs, ensembled_obj_params =\
            self._process_ensembles(params, ensemble_mask)

        # In mixed case, merge with non_ensemble
        ensembled_objs += non_ensemble_objs
        ensembled_obj_params.update(non_ensemble_obj_params)
        return ensembled_objs, ensembled_obj_params

    def _process_ensembles(self, params, ensemble_mask):

        # Separate the ensemble params from all passed params
        ensemble_params = [i for idx, i in enumerate(params)
                           if ensemble_mask[idx]]

        # Get base ensemble objs. and obj_params
        ensembles =\
            EnsembleConstructor(self.user_passed_objs, self.dataset, self.spec)
        ensemble_objs, ensemble_obj_params = ensembles.process(ensemble_params)

        # For each ensemble, go through and process
        ensembled_objs, ensembled_obj_params = [], {}
        for i in range(len(ensemble_params)):

            # Recursively process the base ensembles models
            model_objs, model_obj_params =\
                self.process(ensemble_params[i].models)

            # Check for a base model - process if found
            if ensemble_params[i].base_model is not None:
                final_estimator, final_estimator_params =\
                    self.process(ensemble_params[i].base_model)
            else:
                final_estimator, final_estimator_params = None, {}

            # If any passed ensemble params, then need to select just the
            # params associated with this ensemble
            this_ensemble_name = ensemble_objs[i][0]
            this_ensemble_obj_params =\
                {key: ensemble_obj_params[key] for key
                 in ensemble_obj_params
                 if this_ensemble_name == key.split('__')[0]}

            # Create wrapper object for building the ensemble
            wrapper = EnsembleWrapper(model_obj_params,
                                      this_ensemble_obj_params,
                                      ensembles._get_base_ensembler,
                                      n_jobs=self.spec['n_jobs'],
                                      random_state=self.spec['random_state'])

            # Use the wrapper to get the ensembled_obj
            ensembled_objs +=\
                wrapper.wrap_ensemble(
                    models=model_objs,
                    ensemble=ensemble_objs[i],
                    ensemble_params=ensemble_params[i],
                    final_estimator=final_estimator,
                    final_estimator_params=final_estimator_params
                    )

            # The wrapper keeps track of the changes to params,
            # Add the final updated params to the collective
            # ensembled_obj_params
            ensembled_obj_params.update(wrapper.get_updated_params())

        # Perform wrap checks on the ensembles
        ensembled_objs, ensembled_obj_params =\
            self.check_wraps(ensembled_objs,
                             ensembled_obj_params,
                             ensemble_params)

        return ensembled_objs, ensembled_obj_params

    def check_wraps(self, objs, obj_params, params):
        '''Check for first target scaler wrap, then nested param search,
        then scope model wrap.'''

        # Check ensembled_objs for target_scaler
        for i in range(len(params)):
            target_scaler = params[i].target_scaler

            objs[i], obj_params =\
                self.wrap_target_scaler(target_scaler, objs[i], obj_params)

        # Check for nested param search
        for i in range(len(params)):

            param_search = params[i].param_search

            objs[i], obj_params =\
                wrap_param_search(param_search, objs[i], obj_params)

        # For now am wrapping all models in Scope wrap
        # Could / maybe should change this in the future.
        for i in range(len(params)):
            scope = params[i].scope
            objs[i], obj_params =\
                self.wrap_bpt_model(scope, objs[i], obj_params)

        return objs, obj_params

    def wrap_target_scaler(self, target_scaler, model_obj, model_params):

        if self.spec['problem_type'] != 'regression':
            return model_obj, model_params
        if target_scaler is None:
            return model_obj, model_params

        # Process and get the base scaler_obj + params
        base_scaler_obj = TargetScalerConstructor(self.user_passed_objs,
                                                  self.dataset,
                                                  self.spec)
        scaler_objs, scaler_params =\
            base_scaler_obj.process(target_scaler)
        scaler_obj = scaler_objs[0]

        # Unwrap into name + base
        model_name, base_model = model_obj[0], model_obj[1]
        scaler_name, base_scaler = scaler_obj[0], scaler_obj[1]

        # Now, wrap the model + scaler in the transformed target regressor
        base_wrapper_model =\
            TransformedTargetRegressor(regressor=base_model,
                                       transformer=base_scaler)
        wrapped_name = 'scale_target_' + model_name
        wrapper_model_obj = (wrapped_name, base_wrapper_model)

        # Need to update model params with new nested model name
        model_param_names = list(model_params)
        for param_name in model_param_names:
            if param_name.startswith(model_name + '__'):

                new_base = wrapped_name + '__regressor__'
                new_param_name =\
                    param_name.replace(model_name + '__', new_base, 1)

                model_params[new_param_name] = model_params.pop(param_name)

        # Need to also update / add any scaler params
        for param_name in scaler_params:
            if param_name.startswith(scaler_name + '__'):

                new_base = wrapped_name + '__transformer__'
                new_param_name =\
                    param_name.replace(scaler_name + '__', new_base, 1)

                model_params[new_param_name] = scaler_params[param_name]

        return wrapper_model_obj, model_params

    def wrap_bpt_model(self, scope, model, model_params):

        # Gets model inds, if scope is all
        # gets inds as Ellipsis, otherwise gets as
        # actual inds.
        inds = self._check_scope_all(scope)

        # Get scope name
        scope_name = get_scope_name(scope)
        name = model[0]

        # Get bpt model - under same name as model
        scope_model = (name + scope_name,
                       BPtModel(estimator=model[1], inds=inds))

        # Change any associated params with this model obj
        param_keys = list(model_params)
        for key in param_keys:
            split_key = key.split('__')

            # If this is an associated param w/ this obj
            if split_key[0] == name:

                # Replace name with estimator
                name = split_key[0] + scope_name
                split_key[0] = 'estimator'

                # Create new name
                new_name = '__'.join([name] + split_key)

                # Replace under new name
                model_params[new_name] = model_params.pop(key)

        return scope_model, model_params


class LoaderConstructor(Constructor):

    name = 'loaders'

    def _process_base(self, params):

        from ..default.options.loaders import get_loader_and_params

        # Check for pipe first
        pipe_mask = np.array([isinstance(param.obj, Pipe) for param in params])

        # Init objs as empty list of correct size w/ just place holders
        objs = [None for i in range(len(params))]

        # Process everything but the select groups first
        non_pipe_params = [i for idx, i in enumerate(params)
                           if not pipe_mask[idx]]

        # obj_params has no order, so init with non pipe params
        non_pipe_objs, obj_params =\
            self._get_objs_and_params(get_loader_and_params, non_pipe_params)

        # Update right spot in objs
        for obj, ind in zip(non_pipe_objs, np.where(~pipe_mask)[0]):
            objs[ind] = obj

        # Split up pipe params
        pipe_params = [i for idx, i in enumerate(params) if pipe_mask[idx]]

        # Next, process each group of pipe params seperately
        cnt = 0
        for p_params, ind in zip(pipe_params, np.where(pipe_mask)[0]):

            # Need to move objs and params to separate objs to pass along
            p_sep_params = []
            for o, p in zip(p_params.obj, p_params.params):
                base = deepcopy(p_params)
                base.obj = o
                base.params = p
                p_sep_params.append(base)

            p_objs, p_obj_params =\
                self._get_objs_and_params(get_loader_and_params, p_sep_params)

            # Create loader pipeline
            name = 'loader_pipe' + str(cnt)
            p_obj = (name, Pipeline(steps=p_objs))

            # Add pipeline name to params
            with_name =\
                {name + '__' + key: p_obj_params[key] for key in p_obj_params}

            # Can update final params
            objs[ind] = p_obj
            obj_params.update(with_name)
            cnt += 1

        return objs, obj_params

    def _process(self, params):

        from .BPtLoader import BPtLoader, BPtListLoader

        # Process according to passed tuples or not
        passed_loaders, passed_loader_params =\
            self._process_base(params)

        loaders = []
        for named_obj, param in zip(passed_loaders, params):

            # Get inds from scope
            # special case for loader, don't want
            # to pass Ellipsis for now.
            inds = self.get_inds(param.scope)

            # Unpack to name and object
            name, obj = named_obj

            # For Loader, prioritize wrapper n_jobs
            # so set base obj to 1 if there
            if hasattr(obj, 'n_jobs'):
                setattr(obj, 'n_jobs', 1)

            # Propegate random state
            if hasattr(obj, 'random_state'):
                setattr(obj, 'random_state', self.spec['random_state'])

            # Determine which loader object to use
            if param.behav == 'single':
                Loader = BPtLoader
            else:
                Loader = BPtListLoader

            # Wrap in BPtLoader
            wrapped_obj =\
                Loader(estimator=obj, inds=inds,
                       file_mapping=self.dataset.get_file_mapping(),
                       n_jobs=self.spec['n_jobs'],
                       fix_n_jobs=param.fix_n_wrapper_jobs,
                       cache_loc=param.cache_loc)

            # Add to loaders, use same as base name
            loaders.append((name, wrapped_obj))

        # Process the params, by adding estimator
        loader_params = add_estimator_to_params(passed_loader_params)

        return loaders, loader_params


class ImputerConstructor(Constructor):

    name = 'imputers'

    def _process(self, params):

        from ..default.options.imputers import get_imputer_and_params

        # Then call get objs and params
        objs, obj_params =\
            self._get_objs_and_params(get_imputer_and_params,
                                      params)

        # Replace any base estimators
        objs, obj_params =\
            self.replace_base_estimator(objs, obj_params, params)

        # Make col version
        return self._make_col_version(objs, obj_params, params,
                                      Wrapper=ScopeTransformer)


class TargetScalerConstructor(Constructor):

    name = 'target scalers'

    def _process(self, params):

        from ..default.options.scalers import get_scaler_and_params

        # Then call get objs and params
        objs, obj_params =\
            self._get_objs_and_params(get_scaler_and_params,
                                      params)

        return objs, obj_params


class ScalerConstructor(Constructor):

    name = 'scalers'

    def _process(self, params):

        from ..default.options.scalers import get_scaler_and_params

        # Then call get objs and params
        objs, obj_params =\
            self._get_objs_and_params(get_scaler_and_params,
                                      params)

        # Wrap
        return self._make_col_version(objs, obj_params, params,
                                      Wrapper=ScopeTransformer)


class TransformerConstructor(Constructor):

    name = 'transformers'

    def _process(self, params):

        from ..default.options.transformers import get_transformer_and_params
        from .BPtTransformer import BPtTransformer

        # Then call get objs and params
        objs, obj_params =\
            self._get_objs_and_params(get_transformer_and_params,
                                      params)

        return self._make_col_version(objs, obj_params,
                                      params, Wrapper=BPtTransformer)


class FeatSelectorConstructor(TypeConstructor):

    name = 'feat_selectors'

    def _process(self, params):

        from ..default.options.feature_selectors import (
            get_feat_selector_and_params, AVALIABLE)
        self.AVAILABLE = AVALIABLE

        # Standard base type process
        objs, obj_params =\
            self._base_type_process(params, get_feat_selector_and_params)

        # If any base estimators, replace with a model
        objs, obj_params =\
            self.replace_base_estimator(objs, obj_params, params)

        # Wrap in BPtFeatureSelector
        return self._make_col_version(objs, obj_params, params,
                                      Wrapper=BPtFeatureSelector)


class EnsembleConstructor(TypeConstructor):

    name = 'ensembles'

    def _process(self, params):
        from ..default.options.ensembles import (get_ensemble_and_params,
                                                 AVALIABLE)
        self.AVAILABLE = AVALIABLE

        return self._base_type_process(params,
                                       get_ensemble_and_params)

    def _get_base_ensembler(self, models):

        # @TODO Might want to reflect choice of ensemble / model n_jobs here?

        # If wrapping in ensemble, set n_jobs for ensemble
        # and each indv model, make sure 1
        for model in models:
            try:
                model[1].n_jobs = 1
            except AttributeError:
                pass

            # Ensemble of des ensembles case
            if hasattr(model[1], 'estimators'):
                for estimator in model[1].estimators:
                    try:
                        estimator.n_jobs = 1
                    except AttributeError:
                        pass

        if self.spec['problem_type'] == 'regression':
            return VotingRegressor(models, n_jobs=self.spec['n_jobs'])

        return VotingClassifier(models, voting='soft',
                                n_jobs=self.spec['n_jobs'])


class CustomConstructor():

    name = 'custom'

    def __init__(self, *args, **kwargs):
        '''Compat.'''
        pass

    def process(self, params):
        '''params is passed as a list with all the custom objects.
        No returned param dists either.'''

        # Extract steps
        objs = [param._get_step() for param in params]
        names = [name for name, _ in objs]
        estimators = [est for _, est in objs]

        # Get counts for each name
        namecount = defaultdict(int)
        for name in names:
            namecount[name] += 1

        # Remove any with unique names
        for k, v in list(namecount.items()):
            if v == 1:
                del namecount[k]

        # Proc new names
        for i in reversed(range(len(estimators))):
            name = names[i]
            if name in namecount:
                names[i] += "-%d" % namecount[name]
                namecount[name] -= 1

        objs = list(zip(names, estimators))

        return objs, {}



from sklearn.base import TransformerMixin, clone
from .helpers import proc_mapping, update_mapping
from sklearn.utils.validation import check_memory
import numpy as np
import warnings

from .base import BPtBase, _get_est_fit_params, _needs, _get_est_trans_params


def _fit_estimator(estimator, X, y=None, **fit_params):

    estimator.fit(X=X, y=y, **fit_params)
    return estimator


class ScopeObj(BPtBase):

    _needs_mapping = True
    _needs_fit_index = True

    # Override
    _required_parameters = ["estimator", "inds"]

    def __init__(self, estimator, inds, passthrough=False, cache_loc=None):

        # Set estimator
        super().__init__(estimator=estimator)

        # These are the index to restrict scope to
        self.inds = inds

        self.passthrough = passthrough

        # This is the optional cache_loc for memory
        self.cache_loc = cache_loc

    def __repr__(self):

        temp_inds = self.inds

        if self.inds is Ellipsis:
            self.inds = 'all'
        else:
            self.inds = 'len(' + str(len(self.inds)) + ')'

        rep = super().__repr__()
        self.inds = temp_inds

        return rep

    def _proc_mapping(self, mapping):

        # Save a copy of the passed mapping
        self.mapping_ = mapping.copy()

        # If any mapping passed, update inds in inds_
        if len(mapping) > 0:
            self.inds_ = proc_mapping(self.inds, mapping)

        # Otherwise, inds are as is
        else:
            self.inds_ = self.inds

        return self

    def _fit(self, X, y=None, **fit_params):

        # Get correct fit function as either with memory
        # caching, or just as is, if no cache loc passed.
        if self.cache_loc is not None:
            memory = check_memory(self.cache_loc)
            _fit_estimator_c = memory.cache(_fit_estimator)
        else:
            _fit_estimator_c = _fit_estimator

        # Fit the estimator
        self.estimator_ = _fit_estimator_c(estimator=self.estimator_,
                                           X=X[:, self.inds_],
                                           y=y, **fit_params)

    def _all_case_update_mappings(self, mapping):

        # In case where the scope is everything

        # Rest inds is empty
        self.rest_inds_ = []

        # So is out_mapping
        self.out_mapping_ = {}

        # The mapping to pass on is just
        # a copy of the original
        return mapping.copy()

    def _update_mappings(self, X, mapping):

        # Check in case of Ellipsis / all case
        if self.inds_ is Ellipsis:
            return self._all_case_update_mappings(mapping)

        # Create a mapping that maps old to new value
        self.out_mapping_ = {}

        # For each ind, set sequential
        for i in range(len(self.inds_)):
            self.out_mapping_[self.inds_[i]] = i

        # Compute rest of inds
        self.rest_inds_ =\
            list(np.setdiff1d(list(range(X.shape[1])),
                 self.inds_, assume_unique=True))

        # Set every i in rest_inds_ to None, as out of scope
        for i in self.rest_inds_:
            self.out_mapping_[i] = None

        # Use the out mapping to create a mapping to pass along
        # to any nested objects.
        # Make copy so that the original mapping doesn't change.
        pass_on_mapping = mapping.copy()
        update_mapping(pass_on_mapping, self.out_mapping_)

        return pass_on_mapping

    def fit(self, X, y=None, mapping=None,
            fit_index=None, **fit_params):

        # Save base dtype of input, and n_features_in
        self.base_dtype_ = X.dtype
        self.n_features_in_ = X.shape[1]

        # Process the passed mapping, sets self.inds_
        if mapping is None:
            mapping = {}

        self._proc_mapping(mapping)

        # If no inds to map, skip by setting estimator to None
        if self.inds_ is not Ellipsis and len(self.inds_) == 0:
            self.estimator_ = None
            return self

        # Get pass on mapping
        # All saves class attributes:
        # out_mapping_ and rest_inds_,
        pass_on_mapping = self._update_mappings(X, mapping)

        # Clone estimator, clears previous fits
        self.estimator_ = clone(self.estimator)

        # Get the correct fit params to pass along
        pass_fit_params =\
            _get_est_fit_params(estimator=self.estimator_,
                                mapping=pass_on_mapping,
                                fit_index=fit_index,
                                other_params=fit_params)

        # Fit actual estimator with or without caching
        self._fit(X=X, y=y, **pass_fit_params)

        # Set as fitted
        self.is_fitted_ = True

        return self


class ScopeTransformer(ScopeObj, TransformerMixin):

    @property
    def _needs_transform_index(self):
        return _needs(self.estimator, '_needs_transform_index',
                      'transform_index', 'transform')

    def fit(self, X, y=None, mapping=None,
            fit_index=None, **fit_params):

        if mapping is None:
            mapping = {}

        # Call parent fit - base shared fit with BPtModel
        super().fit(X, y=y, mapping=mapping,
                    fit_index=fit_index,
                    **fit_params)

        # If skip
        if self.estimator_ is None:
            return self

        # Also skip if original scope was all
        # which means the out_mapping shouldn't change
        # and also no updates to the mapping should be made.
        if self.inds_ is Ellipsis:
            return self

        # Now need to make changes to the original mapping
        # to reflect that the new order is self.inds_ + self.rest_inds_
        # or will be after a transform.

        # Use same out mapping from before, but override the None's
        # with the real new spot
        for c in range(len(self.rest_inds_)):
            ind = self.rest_inds_[c]
            self.out_mapping_[ind] = len(self.inds_) + c

        # Passthrough case
        if self.passthrough:

            # Need to update out_mapping_ to reflect that
            # each ind in self.inds_, now maps also to
            # len(self.inds_) + len(self.rest_inds_)
            t = len(self.inds_) + len(self.rest_inds_)
            for i in range(len(self.inds_)):

                extra = i + t
                loc = self.inds_[i]
                current = self.out_mapping_[loc]

                if not isinstance(current, list):
                    current = [current]

                self.out_mapping_[loc] = current + [extra]

        # Update the original mapping, this is the mapping which
        # will be passed to the next piece of the pipeline
        update_mapping(mapping, self.out_mapping_)

        return self

    def transform(self, X, transform_index=None):

        # If None, pass along as is
        if self.estimator_ is None:
            return X

        # Get transform params
        trans_params = _get_est_trans_params(self.estimator_,
                                             transform_index=transform_index)

        # Get X_trans
        X_trans = self._est_transform(X, **trans_params)

        # Save number of output features after X_trans
        self.n_trans_feats_ = X_trans.shape[1]

        # Passthrough case
        if self.passthrough:

            # If no rest inds, skip
            if len(self.rest_inds_) == 0:
                return np.hstack([X_trans, X[:, self.inds_]])

            return np.hstack([X_trans, X[:, self.rest_inds_],
                              X[:, self.inds_]])

        # Return stacked X_trans with rest inds
        if len(self.rest_inds_) > 0:
            return np.hstack([X_trans, X[:, self.rest_inds_]])

        # No rest inds, return directly
        return X_trans

    def _est_transform(self, X, **trans_params):

        # if self.inds_ is Ellipsis, just selects all
        return self.estimator_.transform(X=X[:, self.inds_], **trans_params)

    def fit_transform(self, X, y=None, mapping=None,
                      fit_index=None, transform_index=None,
                      **fit_params):
        '''Override fit transfrom to pass fit_index to transform_index'''

        return self.fit(
            X=X, y=y, mapping=mapping, fit_index=fit_index,
            **fit_params).transform(X=X, transform_index=fit_index)

    def transform_df(self, df, base_name=None, encoders=None):

        # Important: Operate on a copy of the data frame
        df_copy = df.copy()

        # If None, pass along as is
        if self.estimator_ is None:
            return df

        # Get transfrom index from df
        transform_index = df_copy.index

        # Prepare as numpy array - make sure same as original passed dtype
        X = np.array(df_copy).astype(self.base_dtype_)

        # Transform data
        X_trans = self.transform(X, transform_index=transform_index)

        # Feat names are as is
        feat_names = list(df_copy)

        # Process new names
        new_names = self._proc_new_names(feat_names, base_name,
                                         encoders=encoders)

        # Fill in the new values directly to the passed df_copy
        for i, feat_name in enumerate(new_names):
            df_copy.loc[:, feat_name] = X_trans[:, i]

        # Return by re-ordering the df so that it matches
        # the order of new_names, and only with those included in new_names
        return df_copy.loc[:, new_names]

    def _proc_new_names(self, feat_names, base_name=None, encoders=None):

        # If skip, return passed names as is
        if self.estimator_ is None:
            return feat_names

        # If all, return as is
        if self.inds_ is Ellipsis:
            return feat_names

        # Compute new feature names
        new_names = [feat_names[i] for i in self.inds_] +\
                    [feat_names[i] for i in self.rest_inds_]

        return new_names

    def _remove_old_names(self, feat_names):
        '''Create new feature names for the transformed features.
        This class is used in child classes'''

        # If all, all original feat_names get removed
        # return empty list
        if self.inds_ is Ellipsis:
            return []

        to_remove = set([feat_names[i] for i in self.inds_])
        feat_names = [name for name in feat_names if name not in to_remove]
        return feat_names

    def inverse_transform(self, X, name='transformer'):

        # If None, pass along as is
        if self.estimator_ is None:
            return X

        # Compute reverse inds - if Ellipsis, returns Ellipsis
        reverse_inds = proc_mapping(self.inds_, self.out_mapping_)

        # If no inverse_transformer in base transformer, set to 0
        try:
            X_trans =\
                self.estimator_.inverse_transform(X[:, reverse_inds])
        except AttributeError:
            X_trans = np.zeros((X.shape[0], len(self.inds_)))
            warnings.warn('Passed obj: "' + name + '" has no '
                          'inverse_transform! '
                          'Setting relevant inverse '
                          'feat importances to 0.')

        # Compute reverse of rest inds
        reverse_rest_inds = proc_mapping(self.rest_inds_, self.out_mapping_)

        # Create empty Xt to fill
        all_inds_len = len(self.inds_) + len(self.rest_inds_)
        Xt = np.zeros((X.shape[0], all_inds_len), dtype=X.dtype)

        # Fill in Xt
        Xt[:, self.inds_] = X_trans
        Xt[:, self.rest_inds_] = X[:, reverse_rest_inds]

        return Xt



import os
import numpy as np


class BPtInputMixIn():
    pass


def is_array_like(in_val):
    '''Helper function to check if a value is array like or not.

    Parameters
    -----------
    in_val : anything
        The input to check if array like.

    Returns
    ---------
    is_array_like : bool
        True if array like, False if not.

    Examples
    -----------

    .. ipython:: python

        from BPt.util import is_array_like
        import numpy as np

        is_array_like([1, 2, 3])
        is_array_like('some_string')
        is_array_like(5)
        is_array_like(set([1, 2, 3]))
        is_array_like(np.array([1, 2, 3]))

    '''

    if hasattr(in_val, '__len__') and (not isinstance(in_val, str)) and \
       (not isinstance(in_val, dict)) and (not hasattr(in_val, 'fit')) and \
       (not hasattr(in_val, 'transform')):
        return True
    else:
        return False


def conv_to_list(in_val, amt=1):

    if in_val is None:
        return None

    if not is_array_like(in_val) or isinstance(in_val, BPtInputMixIn):
        in_val = [in_val for i in range(amt)]

    return in_val


def save_docx_table(df, filename, decimals=3):
    '''| Helper function for saving a dataframe to a docx file.

    | You must have the library python-docx installed to use
        this function.

    Parameters
    -----------
    df : pandas DataFrame
        The DataFrame in which to save as a docx table.

    filename : str
        A path to a docx file in which to save the table.
        If the file already exists, then the table will be appended to
        the existing file, if it doesn't already exist,
        then it will be created.

    decimals : int, optional
        An optional number of decimal points to
        round any floating point numbers to in the saved
        table.

        ::

            default = 3

    See Also
    ----------
    BPt.Dataset.summary : Create and optionally save a summary of columns.

    '''

    import docx

    # Reset index
    df = df.reset_index()

    if os.path.exists(filename):
        doc = docx.Document(filename)
    else:
        doc = docx.Document()

    t = doc.add_table(df.shape[0]+1, df.shape[1])

    # For each column
    for j in range(df.shape[-1]):

        # Add header
        t.cell(0, j).text = df.columns[j]

        # Check if column is type is float
        col = list(df)[j]
        is_float = 'float' in df[col].dtype.name.lower()

        for i in range(df.shape[0]):

            # Get value
            value = df.values[i, j]

            # Round if float
            if is_float:
                value = np.round(value, decimals=decimals)

            # Set value
            t.cell(i+1, j).text = str(value)

    doc.save(filename)


def substrs(x):
    return {x[i:i+j] for i in range(len(x)) for j in range(len(x) - i + 1)}


def find_substr(data):

    s = substrs(data[0])

    for val in data[1:]:
        s.intersection_update(substrs(val))

    try:
        mx = max(s, key=len)

    except ValueError:
        mx = ''

    return mx


def get_top_substrs(keys):

    found = []
    top = find_substr(keys)

    while len(top) > 1:
        found.append(top)

        keys = [k.replace(top, '') for k in keys]
        top = find_substr(keys)

    return found


import numpy as np
from ..data_file import DataFile, mp_single_load, load_data_file_proxy
import tempfile
import pandas as pd
import os


def get_fake_data_file():

    temp_dr = tempfile.gettempdir()
    file_loc = os.path.join(temp_dr, 'fake.npy')

    data = np.arange(1, 10)
    np.save(file_loc, data)

    df = DataFile(loc=file_loc, load_func=np.load)

    return df


def test_Data_File():

    df = get_fake_data_file()
    assert df.loc.endswith('fake.npy')

    data = df.load()
    assert len(data) == 9
    assert np.max(data) == 9

    # Make sure equality based on file path works
    df2 = get_fake_data_file()
    assert df == df2

    # Make sure hash works
    assert hash(df) == hash(df2)


def test_Data_File_repr():

    df = get_fake_data_file()

    assert repr(df).startswith("DataFile(loc=")
    assert str(df).startswith("DataFile(loc=")


def test_df_sort():

    df = get_fake_data_file()

    assert not df < df


def test_mp_single_load():

    df = get_fake_data_file()
    files = [df, df]

    proxy = mp_single_load(files=files, reduce_func=np.max)
    assert all(proxy == 9)


def test_load_data_file_proxy():

    df = get_fake_data_file()

    # Fake file mapping
    file_mapping = {}
    for i in range(10):
        file_mapping[i] = df

    values = pd.Series(np.arange(10))

    proxy =\
        load_data_file_proxy(values, reduce_func=np.min,
                             file_mapping=file_mapping, n_jobs=1)
    assert proxy.sum() == 10
    assert proxy[0] == 1
    assert proxy[9] == 1

    # N jobs == 2
    proxy =\
        load_data_file_proxy(values, reduce_func=np.min,
                             file_mapping=file_mapping, n_jobs=2)
    assert proxy.sum() == 10
    assert proxy[0] == 1
    assert proxy[9] == 1



from ..Dataset import Dataset
import numpy as np
import tempfile
import os
import warnings


def get_fake_dataset():

    fake = Dataset()
    fake['class'] = [1, 1, 1, 2, 2, 2]
    fake['class2'] = [2, 2, 2, 1, 1, 1]
    fake['val1'] = [.1, .5, 1, 4, 1, np.nan]
    fake['val2'] = [2, 1, .1, 5, 2, 7]

    fake = fake.add_scope('class', 'category')
    fake = fake.add_scope('class2', 'category')

    return fake


def test_for_no_errors():

    df = get_fake_dataset()
    df.plot('all', show=False)
    df.plot_bivar('val2', 'val1', show=False)
    df.plot_bivar('class', 'class2', show=False)
    df.plot_bivar('class', 'val1', show=False)


def test_summary_cat_col():

    df = get_fake_dataset()

    df1, df2 =\
        df.summary(scope='class',
                   cat_measures=['count', 'freq', 'nan count'],
                   decode_values=True)

    assert len(df1) == 0

    assert df2.loc['class', 'count'] == 6
    assert df2.loc['class=1', 'count'] == 3
    assert df2.loc['class=2', 'count'] == 3

    assert df2.loc['class', 'freq'] == 1
    assert df2.loc['class=1', 'freq'] == .5
    assert df2.loc['class=2', 'freq'] == .5

    assert df2.loc['class', 'nan count'] == 0
    assert df2.loc['class=1', 'nan count'] == 0
    assert df2.loc['class=2', 'nan count'] == 0


def test_summary_float_col_mean_std():

    df = get_fake_dataset()

    df1, df2 =\
        df.summary(scope='val1',
                   measures=['mean +- std'])

    assert len(df2) == 0
    print(df1)
    assert df1.loc['val1', 'mean ± std'] == '1.32 ± 1.545'


def test_summary_float_col():

    df = get_fake_dataset()

    df1, df2 =\
        df.summary(scope='val1',
                   measures=['count', 'nan count',
                             'mean', 'max',
                             'min', 'std', 'var',
                             'skew', 'kurtosis'])

    assert len(df2) == 0

    # Loose tests, more want to make sure results are there.
    assert df1.loc['val1', 'count'] == 5
    assert df1.loc['val1', 'nan count'] == 1
    assert df1.loc['val1', 'mean'] > 1.31
    assert df1.loc['val1', 'mean'] < 1.33
    assert df1.loc['val1', 'max'] == 4
    assert df1.loc['val1', 'min'] == .1
    assert df1.loc['val1', 'std'] > 1.5
    assert df1.loc['val1', 'std'] < 1.6
    assert df1.loc['val1', 'skew'] < 1.9
    assert df1.loc['val1', 'skew'] > 1.88
    assert df1.loc['val1', 'kurtosis'] > 3.9
    assert df1.loc['val1', 'kurtosis'] < 3.91
    assert df1.loc['val1', 'var'] < 2.388
    assert df1.loc['val1', 'var'] > 2.386


def test_summary_save():

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        df = get_fake_dataset()

        temp = os.path.join(tempfile.gettempdir(), 't1.docx')
        if os.path.exists(temp):
            os.remove(temp)

        df.summary(scope=['class', 'val1'],
                   save_file=temp)

        assert os.path.exists(temp)
        os.remove(temp)



from sklearn.metrics import SCORERS
import sklearn.metrics as M
from ..helpers import proc_type_dep_str
from ...util import conv_to_list
from joblib import wrap_non_picklable_objects

AVALIABLE = {
    'binary': {
        'accuracy': 'accuracy',
        'roc_auc': 'roc_auc',
        'roc_auc_ovr': 'roc_auc_ovr',
        'roc_auc_ovo': 'roc_auc_ovo',
        'roc_auc_ovr_weighted': 'roc_auc_ovr_weighted',
        'roc_auc_ovo_weighted': 'roc_auc_ovo_weighted',
        'balanced_accuracy': 'balanced_accuracy',
        'average_precision': 'average_precision',
        'neg_log_loss': 'neg_log_loss',
        'neg_brier_score': 'neg_brier_score',
        'precision': 'precision',
        'precision_macro': 'precision_macro',
        'precision_micro': 'precision_micro',
        'precision_samples': 'precision_samples',
        'precision_weighted': 'precision_weighted',
        'recall': 'recall',
        'recall_macro': 'recall_macro',
        'recall_micro': 'recall_micro',
        'recall_samples': 'recall_samples',
        'recall_weighted': 'recall_weighted',
        'f1': 'f1',
        'f1_macro': 'f1_macro',
        'f1_micro': 'f1_micro',
        'f1_samples': 'f1_samples',
        'f1_weighted': 'f1_weighted',
        'jaccard': 'jaccard',
        'jaccard_macro': 'jaccard_macro',
        'jaccard_micro': 'jaccard_micro',
        'jaccard_samples': 'jaccard_samples',
        'jaccard_weighted': 'jaccard_weighted',
        'neg_hamming': 'neg_hamming',
        'matthews': 'matthews',
    },
    'regression': {
        'explained_variance': 'explained_variance',
        'explained_variance score': 'explained_variance',
        'r2': 'r2',
        'max_error': 'max_error',
        'neg_median_absolute_error': 'neg_median_absolute_error',
        'median_absolute_error': 'neg_median_absolute_error',
        'neg_mean_absolute_error': 'neg_mean_absolute_error',
        'mean_absolute_error': 'neg_mean_absolute_error',
        'neg_mean_squared_error': 'neg_mean_squared_error',
        'mean_squared_error': 'neg_mean_squared_error',
        'neg_mean_squared_log_error': 'neg_mean_squared_log_error',
        'mean_squared_log_error': 'neg_mean_squared_log_error',
        'neg_root_mean_squared_error': 'neg_root_mean_squared_error',
        'root_mean_squared_error': 'neg_root_mean_squared_error',
        'neg_mean_poisson_deviance': 'neg_mean_poisson_deviance',
        'mean_poisson_deviance': 'neg_mean_poisson_deviance',
        'neg_mean_gamma_deviance': 'neg_mean_gamma_deviance',
        'mean_gamma_deviance': 'neg_mean_gamma_deviance',
    },
}
AVALIABLE['categorical'] = AVALIABLE['binary'].copy()

# Set defaults
AVALIABLE['binary']['default'] = 'roc_auc'
AVALIABLE['regression']['default'] = 'r2'
AVALIABLE['categorical']['default'] = 'roc_auc_ovr'

SCORERS.update({

    'neg_hamming': M.make_scorer(score_func=M.hamming_loss,
                                 greater_is_better=False,
                                 needs_threshold=False),

    'matthews': M.make_scorer(score_func=M.matthews_corrcoef,
                              greater_is_better=True)

})

default_scorers = {'regression': ['explained_variance',
                                  'neg_mean_squared_error'],
                   'binary': ['matthews', 'roc_auc',
                              'balanced_accuracy'],
                   'categorical': ['matthews', 'roc_auc_ovr',
                                   'balanced_accuracy']}


def get_scorer_from_str(scorer_str):

    return SCORERS[scorer_str]


def get_scorers_by_type(problem_type):

    objs = []
    for scorer_str in AVALIABLE[problem_type]:
        conv = AVALIABLE[problem_type][scorer_str]
        score_func = get_scorer_from_str(conv)._score_func
        objs.append((scorer_str, score_func))

    return objs


def process_scorers(in_scorers, problem_type):

    # If already correct scorers dict.
    if isinstance(in_scorers, dict):
        return in_scorers

    # Check for default case
    if in_scorers == 'default':
        in_scorers = default_scorers[problem_type]

    # get scorer_strs as initially list
    scorer_strs = conv_to_list(in_scorers)

    scorers, cnt = {}, 0
    for m in range(len(scorer_strs)):

        if isinstance(scorer_strs[m], str):
            name = proc_type_dep_str(scorer_strs[m], AVALIABLE,
                                     problem_type)
            scorers[name] = get_scorer_from_str(name)

        else:
            name = 'Custom Scorer ' + str(cnt)
            scorers[name] = wrap_non_picklable_objects(scorer_strs[m])
            cnt += 1

    return scorers


def process_scorer(scorer_str, problem_type):

    # Process scorer strs if default
    if scorer_str == 'default':
        scorer_str = default_scorers[problem_type][0]

    # If passed as str
    if isinstance(scorer_str, str):
        scorer_str = proc_type_dep_str(scorer_str, AVALIABLE, problem_type)
        return get_scorer_from_str(scorer_str)

    # Otherwise assume function and wrap
    return wrap_non_picklable_objects(scorer_str)
